#  Copyright (c) 2022-2024, Manfred Moitzi
#  License: MIT License
from __future__ import annotations
from typing import TypeVar, Generic, TYPE_CHECKING, Optional
from abc import ABC, abstractmethod
from .const import NULL_PTR_NAME, MIN_EXPORT_VERSION
from .hdr import AcisHeader

if TYPE_CHECKING:
    from .entities import AcisEntity
    from ezdxf.math import Vec3


T = TypeVar("T", bound="AbstractEntity")


class AbstractEntity(ABC):
    """Unified query interface for SAT and SAB data."""

    name: str
    id: int = -1

    def __str__(self):
        return f"{self.name}"

    @property
    def is_null_ptr(self) -> bool:
        """Returns ``True`` if this entity is the ``NULL_PTR`` entity."""
        return self.name == NULL_PTR_NAME


class DataLoader(ABC):
    """Data loading interface to create high level AcisEntity data from low
    level AbstractEntity representation.

    """

    version: int = MIN_EXPORT_VERSION

    @abstractmethod
    def has_data(self) -> bool:
        pass

    @abstractmethod
    def read_int(self, skip_sat: Optional[int] = None) -> int:
        """There are sometimes additional int values in SAB files which are
        not present in SAT files, maybe reference counters e.g. vertex, coedge.
        """
        pass

    @abstractmethod
    def read_double(self) -> float:
        pass

    @abstractmethod
    def read_interval(self) -> float:
        pass

    @abstractmethod
    def read_vec3(self) -> tuple[float, float, float]:
        pass

    @abstractmethod
    def read_bool(self, true: str, false: str) -> bool:
        pass

    @abstractmethod
    def read_str(self) -> str:
        pass

    @abstractmethod
    def read_ptr(self) -> AbstractEntity:
        pass

    @abstractmethod
    def read_transform(self) -> list[float]:
        pass


class DataExporter(ABC):
    version: int = MIN_EXPORT_VERSION

    @abstractmethod
    def write_int(self, value: int, skip_sat=False) -> None:
        """There are sometimes additional int values in SAB files which are
        not present in SAT files, maybe reference counters e.g. vertex, coedge.
        """
        pass

    @abstractmethod
    def write_double(self, value: float) -> None:
        pass

    @abstractmethod
    def write_interval(self, value: float) -> None:
        pass

    @abstractmethod
    def write_loc_vec3(self, value: Vec3) -> None:
        pass

    @abstractmethod
    def write_dir_vec3(self, value: Vec3) -> None:
        pass

    @abstractmethod
    def write_bool(self, value: bool, true: str, false: str) -> None:
        pass

    @abstractmethod
    def write_str(self, value: str) -> None:
        pass

    @abstractmethod
    def write_literal_str(self, value: str) -> None:
        pass

    @abstractmethod
    def write_ptr(self, entity: AcisEntity) -> None:
        pass

    @abstractmethod
    def write_transform(self, data: list[str]) -> None:
        pass


class AbstractBuilder(Generic[T]):
    header: AcisHeader
    bodies: list[T]
    entities: list[T]

    def reorder_records(self) -> None:
        if len(self.entities) == 0:
            return
        header: list[T] = []
        entities: list[T] = []
        for e in self.entities:
            if e.name == "body":
                header.append(e)
            elif e.name == "asmheader":
                header.insert(0, e)  # has to be the first record
            else:
                entities.append(e)
        self.entities = header + entities

    def reset_ids(self, start: int = 0) -> None:
        for num, entity in enumerate(self.entities, start=start):
            entity.id = num

    def clear_ids(self) -> None:
        for entity in self.entities:
            entity.id = -1


class EntityExporter(Generic[T]):
    def __init__(self, header: AcisHeader):
        self.header = header
        self.version = header.version
        self._exported_entities: dict[int, T] = {}
        if self.header.has_asm_header:
            self.export(self.header.asm_header())

    def export_records(self) -> list[T]:
        return list(self._exported_entities.values())

    @abstractmethod
    def make_record(self, entity: AcisEntity) -> T:
        pass

    @abstractmethod
    def make_data_exporter(self, record: T) -> DataExporter:
        pass

    def get_record(self, entity: AcisEntity) -> T:
        assert not entity.is_none
        return self._exported_entities[id(entity)]

    def export(self, entity: AcisEntity):
        if entity.is_none:
            raise TypeError("invalid NONE_REF entity given")
        self._make_all_records(entity)
        self._export_data(entity)

    def _has_record(self, entity: AcisEntity) -> bool:
        return id(entity) in self._exported_entities

    def _add_record(self, entity: AcisEntity, record: T) -> None:
        assert not entity.is_none
        self._exported_entities[id(entity)] = record

    def _make_all_records(self, entity: AcisEntity):
        def add(e: AcisEntity) -> bool:
            if not e.is_none and not self._has_record(e):
                self._add_record(e, self.make_record(e))
                return True
            return False

        entities = [entity]
        while entities:
            next_entity = entities.pop(0)
            add(next_entity)
            for sub_entity in next_entity.entities():
                if add(sub_entity):
                    entities.append(sub_entity)

    def _export_data(self, entity: AcisEntity):
        def _export_record(e: AcisEntity):
            if id(e) not in done:
                done.add(id(e))
                record = self.get_record(e)
                if not e.attributes.is_none:
                    record.attributes = self.get_record(e.attributes)  # type: ignore
                e.export(self.make_data_exporter(record))
                return True
            return False

        entities = [entity]
        done: set[int] = set()
        while entities:
            next_entity = entities.pop(0)
            _export_record(next_entity)
            for sub_entity in next_entity.entities():
                if _export_record(sub_entity):
                    entities.append(sub_entity)



# Copyright (c) 2018-2020, Manfred Moitzi
# License: MIT License
import pytest
import ezdxf
from ezdxf import recover
import os

CP936_FILE = os.path.join(
    ezdxf.EZDXF_TEST_FILES, "ChineseChars_cp936_R2004.dxf"
)


@pytest.mark.skipif(
    not os.path.exists(CP936_FILE), reason=f"File {CP936_FILE} not found"
)
def test_read_plain_cp936_chinese():
    doc = ezdxf.readfile(CP936_FILE)
    assert doc.filename is not None
    assert doc.dxfversion is not None


CP936_ZIP_FILE = os.path.join(
    ezdxf.EZDXF_TEST_FILES, "ChineseChars_cp936_R2004.zip"
)


@pytest.mark.skipif(
    not os.path.exists(CP936_ZIP_FILE),
    reason=f"File {CP936_ZIP_FILE} not found",
)
def test_read_from_zip():
    doc = ezdxf.readzip(CP936_ZIP_FILE)
    assert doc.filename is not None
    assert doc.dxfversion is not None


PROE_FILE = os.path.join(ezdxf.EZDXF_TEST_FILES, "ProE_AC1018.dxf")


@pytest.mark.skipif(
    not os.path.exists(PROE_FILE), reason=f'File "{PROE_FILE}" not found'
)
class TestProE:
    @pytest.fixture
    def filename(self):
        return PROE_FILE

    def test_read_ProE_file(self, filename):
        doc = ezdxf.readfile(filename)
        assert doc.filename is not None
        assert doc.dxfversion is not None

    def test_recover_ProE_file(self, filename):
        doc, auditor = recover.readfile(filename)
        assert doc.filename == filename
        assert doc.dxfversion is not None
        assert len(auditor.errors) >= 2
        assert len(auditor.fixes) >= 9


FILE_CIVIL_3D = os.path.join(
    ezdxf.EZDXF_TEST_FILES, "AutodeskProducts/Civil3D_2018.dxf"
)


@pytest.mark.skipif(
    not os.path.exists(FILE_CIVIL_3D),
    reason=f'File "{FILE_CIVIL_3D}" not found',
)
class TestCorruptCivil3D:
    """The test file "AutodeskProducts/Civil3D_2018.dxf" contains binary data
    in XRECORDS and is not valid for TrueView 2020!

    """

    @pytest.fixture
    def filename(self):
        return FILE_CIVIL_3D

    def test_readfile_and_preserve_errors_by_surrogate_escape(self, filename):
        doc = ezdxf.readfile(filename, errors="surrogateescape")
        assert doc.filename == filename
        assert doc.dxfversion == "AC1032"

    def test_readfile_and_ignore_errors(self, filename):
        doc = ezdxf.readfile(filename, errors="ignore")
        assert doc.filename == filename
        assert doc.dxfversion == "AC1032"

    def test_readfile_and_raise_exception(self, filename):
        with pytest.raises(UnicodeDecodeError):
            ezdxf.readfile(filename, errors="strict")

        with pytest.raises(UnicodeDecodeError):
            recover.readfile(filename, errors="strict")

    def test_recover(self, filename):
        doc, auditor = recover.readfile(filename, errors="ignore")
        assert doc.filename == filename
        assert doc.dxfversion == "AC1032"
        assert auditor.has_errors is True
        assert len(auditor.errors) == 238, "expected decoding errors"
        assert len(auditor.fixes) == 536


FILE_MAP_3D = os.path.join(
    ezdxf.EZDXF_TEST_FILES, "AutodeskProducts/Map3D_2017.dxf"
)


@pytest.mark.skipif(
    not os.path.exists(FILE_MAP_3D), reason=f'File "{FILE_MAP_3D}" not found'
)
def test_read_map_3d():
    doc = ezdxf.readfile(FILE_MAP_3D)
    assert doc.filename == FILE_MAP_3D
    assert doc.dxfversion == "AC1027"


GERBER_FILE = os.path.join(ezdxf.EZDXF_TEST_FILES, "Gerber.dxf")


@pytest.mark.skipif(
    not os.path.exists(GERBER_FILE), reason=f'File "{GERBER_FILE}" not found'
)
def test_read_gerber_file():
    from ezdxf.lldxf.validator import is_dxf_file

    assert is_dxf_file(GERBER_FILE)

    doc = ezdxf.readfile(GERBER_FILE)
    assert doc.filename == GERBER_FILE
    assert doc.dxfversion == "AC1009"



# Copyright (c) 2024, Manfred Moitzi
# License: MIT License
from __future__ import annotations
from typing import Iterable, Sequence
import dataclasses

from ezdxf.lldxf import const
from ezdxf.lldxf.tags import Tags
from ezdxf.lldxf.types import dxftag
from ezdxf.entities import SpatialFilter, DXFEntity, Dictionary, Insert, XRecord
from ezdxf.math import Vec2, Vec3, UVec, Z_AXIS, Matrix44, BoundingBox2d
from ezdxf.entities.acad_xrec_roundtrip import RoundtripXRecord

__all__ = ["get_spatial_filter", "XClip", "ClippingPath"]

ACAD_FILTER = "ACAD_FILTER"
ACAD_XREC_ROUNDTRIP = "ACAD_XREC_ROUNDTRIP"
ACAD_INVERTEDCLIP_ROUNDTRIP = "ACAD_INVERTEDCLIP_ROUNDTRIP"
ACAD_INVERTEDCLIP_ROUNDTRIP_COMPARE = "ACAD_INVERTEDCLIP_ROUNDTRIP_COMPARE"

SPATIAL = "SPATIAL"


@dataclasses.dataclass
class ClippingPath:
    """Stores the SPATIAL_FILTER clipping paths in original form that I still don't fully
    understand for `inverted` clipping paths. All boundary paths are simple polygons as a
    sequence of :class:`~ezdxf.math.Vec2`.

    Attributes:
        vertices: Contains the boundary polygon for regular clipping paths.
            Contains the outer boundary path for inverted clippings paths - but not always!
        inverted_clip:
            Contains the inner boundary for inverted clipping paths - but not always!
        inverted_clip_compare:
            Contains the combined inner- and the outer boundaries for inverted
            clipping paths - but not always!
        is_inverted_clip: ``True`` for inverted clipping paths

    """

    vertices: Sequence[Vec2] = tuple()
    inverted_clip: Sequence[Vec2] = tuple()
    inverted_clip_compare: Sequence[Vec2] = tuple()
    is_inverted_clip: bool = False

    def inner_polygon(self) -> Sequence[Vec2]:
        """Returns the inner clipping polygon as sequence of Vec2."""
        # The exact data structure of inverted clippings polygons is still not
        # clear to me, so use the smallest polygon as inner clipping polygon.
        if not self.is_inverted_clip:
            return self.vertices
        inner_polygon = self.vertices
        if bbox_area(self.inverted_clip) < bbox_area(inner_polygon):
            inner_polygon = self.inverted_clip
        return inner_polygon

    def outer_bounds(self) -> BoundingBox2d:
        """Returns the maximum extents as BoundingBox2d."""
        if not self.is_inverted_clip:
            return BoundingBox2d(self.vertices)
        # The exact data structure of inverted clippings polygons is still not
        # clear to me, this is my best guess:
        return BoundingBox2d(self.inverted_clip_compare)


class XClip:
    """Helper class to manage the clipping path of INSERT entities.

    Provides a similar functionality as the XCLIP command in CAD applications.

    .. important::

        This class handles only 2D clipping paths.

    The visibility of the clipping path can be set individually for each block
    reference, but the HEADER variable $XCLIPFRAME ultimately determines whether the
    clipping path is displayed or plotted by the application:

    === =============== ===
    0   not displayed   not plotted
    1   displayed       not plotted
    2   displayed       plotted
    === =============== ===

    The default setting is 2.

    """

    def __init__(self, insert: Insert) -> None:
        if not isinstance(insert, Insert):
            raise const.DXFTypeError(f"INSERT entity required, got {str(insert)}")
        self._insert = insert
        self._spatial_filter = get_spatial_filter(insert)

    def get_spatial_filter(self) -> SpatialFilter | None:
        """Returns the underlaying SPATIAL_FILTER entity if the INSERT entity has a
        clipping path and returns ``None`` otherwise.
        """
        return self._spatial_filter

    def get_xclip_frame_policy(self) -> int:
        policy: int = 2
        if self._insert.doc is not None:
            policy = self._insert.doc.header.get("$XCLIPFRAME", 2)
        return policy

    @property
    def has_clipping_path(self) -> bool:
        """Returns if the INSERT entity has a clipping path."""
        return self._spatial_filter is not None

    @property
    def is_clipping_enabled(self) -> bool:
        """Returns ``True`` if block reference clipping is enabled."""
        if isinstance(self._spatial_filter, SpatialFilter):
            return bool(self._spatial_filter.dxf.is_clipping_enabled)
        return False

    @property
    def is_inverted_clip(self) -> bool:
        """Returns ``True`` if clipping path is inverted."""
        xrec = get_roundtrip_xrecord(self._spatial_filter)
        if xrec is None:
            return False
        return xrec.has_section(ACAD_INVERTEDCLIP_ROUNDTRIP)

    def enable_clipping(self) -> None:
        """Enable block reference clipping."""
        if isinstance(self._spatial_filter, SpatialFilter):
            self._spatial_filter.dxf.is_clipping_enabled = 1

    def disable_clipping(self) -> None:
        """Disable block reference clipping."""
        if isinstance(self._spatial_filter, SpatialFilter):
            self._spatial_filter.dxf.is_clipping_enabled = 0

    def get_block_clipping_path(self) -> ClippingPath:
        """Returns the clipping path in block coordinates (relative to the block origin)."""
        vertices: Sequence[Vec2] = []
        if not isinstance(self._spatial_filter, SpatialFilter):
            return ClippingPath()
        m = self._spatial_filter.inverse_insert_matrix
        vertices = Vec2.tuple(
            m.transform_vertices(self._spatial_filter.boundary_vertices)
        )
        if len(vertices) == 2:
            vertices = _rect_path(vertices)

        clipping_path = ClippingPath(vertices, is_inverted_clip=False)
        xrec = get_roundtrip_xrecord(self._spatial_filter)
        if isinstance(xrec, RoundtripXRecord):
            clipping_path.inverted_clip = get_roundtrip_vertices(
                xrec, ACAD_INVERTEDCLIP_ROUNDTRIP, m
            )
            clipping_path.inverted_clip_compare = get_roundtrip_vertices(
                xrec, ACAD_INVERTEDCLIP_ROUNDTRIP_COMPARE, m
            )
            clipping_path.is_inverted_clip = True
        return clipping_path

    def get_wcs_clipping_path(self) -> ClippingPath:
        """Returns the clipping path in WCS coordinates (relative to the WCS origin) as
        2D path projected onto the xy-plane.
        """
        vertices: Sequence[Vec2] = tuple()
        if not isinstance(self._spatial_filter, SpatialFilter):
            return ClippingPath(vertices, vertices)
        block_clipping_path = self.get_block_clipping_path()
        m = self._insert.matrix44()
        vertices = Vec2.tuple(m.transform_vertices(block_clipping_path.vertices))
        if len(vertices) == 2:  # rectangle by diagonal corner vertices
            vertices = BoundingBox2d(vertices).rect_vertices()
        wcs_clipping_path = ClippingPath(
            vertices, is_inverted_clip=block_clipping_path.is_inverted_clip
        )
        if block_clipping_path.is_inverted_clip:
            inverted_clip = Vec2.tuple(
                m.transform_vertices(block_clipping_path.inverted_clip)
            )
            if len(inverted_clip) == 2:  # rectangle by diagonal corner vertices
                inverted_clip = BoundingBox2d(inverted_clip).rect_vertices()
            wcs_clipping_path.inverted_clip = inverted_clip
            wcs_clipping_path.inverted_clip_compare = Vec2.tuple(
                m.transform_vertices(block_clipping_path.inverted_clip_compare)
            )
        return wcs_clipping_path

    def set_block_clipping_path(self, vertices: Iterable[UVec]) -> None:
        """Set clipping path in block coordinates (relative to block origin).

        The clipping path is located in the xy-plane, the z-axis of all vertices will
        be ignored.  The clipping path doesn't have to be closed (first vertex != last vertex).
        Two vertices define a rectangle where the sides are parallel to x- and y-axis.

        Raises:
           DXFValueError: clipping path has less than two vertrices

        """
        if self._spatial_filter is None:
            self._spatial_filter = new_spatial_filter(self._insert)
        spatial_filter = self._spatial_filter
        spatial_filter.set_boundary_vertices(vertices)
        spatial_filter.dxf.origin = Vec3(0, 0, 0)
        spatial_filter.dxf.extrusion = Z_AXIS
        spatial_filter.dxf.has_front_clipping_plane = 0
        spatial_filter.dxf.front_clipping_plane_distance = 0.0
        spatial_filter.dxf.has_back_clipping_plane = 0
        spatial_filter.dxf.back_clipping_plane_distance = 0.0

        # The clipping path set by ezdxf is always relative to the block origin and
        # therefore both transformation matrices are the identity matrix - which does
        # nothing.
        m = Matrix44()
        spatial_filter.set_inverse_insert_matrix(m)
        spatial_filter.set_transform_matrix(m)
        self._discard_inverted_clip()

    def set_wcs_clipping_path(self, vertices: Iterable[UVec]) -> None:
        """Set clipping path in WCS coordinates (relative to WCS origin).

        The clipping path is located in the xy-plane, the z-axis of all vertices will
        be ignored. The clipping path doesn't have to be closed (first vertex != last vertex).
        Two vertices define a rectangle where the sides are parallel to x- and y-axis.

        Raises:
           DXFValueError: clipping path has less than two vertrices
           ZeroDivisionError: Block reference transformation matrix is not invertible

        """
        m = self._insert.matrix44()
        try:
            m.inverse()
        except ZeroDivisionError:
            raise ZeroDivisionError(
                "Block reference transformation matrix is not invertible."
            )
        _vertices = Vec2.list(vertices)
        if len(_vertices) == 2:
            _vertices = _rect_path(_vertices)
        self.set_block_clipping_path(m.transform_vertices(_vertices))

    def invert_clipping_path(
        self, extents: Iterable[UVec] | None = None, *, ignore_acad_compatibility=False
    ) -> None:
        """Invert clipping path. (experimental feature)

        The outer boundary is defined by the bounding box of the given `extents`
        vertices or auto-detected if `extents` is ``None``.

        The `extents` are BLOCK coordinates.
        Requires an existing clipping path and that clipping path cannot be inverted.

        .. warning::

            You have to set the flag `ignore_acad_compatibility` to ``True`` to use
            this feature.  AutoCAD will not load DXF files with inverted clipping paths
            created by ezdxf!!!!

        """
        if ignore_acad_compatibility is False:
            return

        current_clipping_path = self.get_block_clipping_path()
        if len(current_clipping_path.vertices) < 2:
            raise const.DXFValueError("no clipping path set")
        if current_clipping_path.is_inverted_clip:
            raise const.DXFValueError("clipping path is already inverted")

        assert self._insert.doc is not None
        self._insert.doc.add_acad_incompatibility_message(
            "\nAutoCAD will not load DXF files with inverted clipping paths created by ezdxf"
        )
        grow_factor = 0.0
        if extents is None:
            extents = self._detect_block_extents()
            # grow bounding box by 10%, bbox detection is not very precise for text
            # based entities:
            grow_factor = 0.1

        bbox = BoundingBox2d(extents)
        bbox.extend(current_clipping_path.vertices)
        if not bbox.has_data:
            raise const.DXFValueError("extents not detectable")

        if grow_factor:
            bbox.grow(max(bbox.size) * grow_factor)

        # inverted_clip is the regular clipping path
        inverted_clip = current_clipping_path.vertices
        # construct an inverted clipping path
        inverted_clip_compare = _get_inverted_clip_compare_vertices(bbox, inverted_clip)
        # set inverted_clip_compare as regular clipping path
        self.set_block_clipping_path(inverted_clip_compare)
        self._set_inverted_clipping_path(inverted_clip, inverted_clip_compare)

    def _detect_block_extents(self) -> Sequence[Vec2]:
        from ezdxf import bbox

        insert = self._insert
        doc = insert.doc
        assert doc is not None, "valid DXF document required"
        no_vertices: Sequence[Vec2] = tuple()
        block = doc.blocks.get(insert.dxf.name)
        if block is None:
            return no_vertices

        _bbox = bbox.extents(block, fast=True)
        if not _bbox.has_data:
            return no_vertices
        return Vec2.tuple([_bbox.extmin, _bbox.extmax])

    def _set_inverted_clipping_path(
        self, clip_vertices: Iterable[Vec2], compare_vertices: Iterable[Vec2]
    ) -> None:
        spatial_filter = self._spatial_filter
        assert isinstance(spatial_filter, SpatialFilter)
        xrec = get_roundtrip_xrecord(spatial_filter)
        if xrec is None:
            xrec = new_roundtrip_xrecord(spatial_filter)

        clip_tags = Tags(dxftag(10, Vec3(v)) for v in clip_vertices)
        compare_tags = Tags(dxftag(10, Vec3(v)) for v in compare_vertices)
        xrec.set_section(ACAD_INVERTEDCLIP_ROUNDTRIP, clip_tags)
        xrec.set_section(ACAD_INVERTEDCLIP_ROUNDTRIP_COMPARE, compare_tags)

    def discard_clipping_path(self) -> None:
        """Delete the clipping path. The clipping path doesn't have to exist.

        This method does not discard the extension dictionary of the base entity,
        even when its empty.
        """
        if not isinstance(self._spatial_filter, SpatialFilter):
            return

        xdict = self._insert.get_extension_dict()
        xdict.discard(ACAD_FILTER)
        entitydb = self._insert.doc.entitydb  # type: ignore
        assert entitydb is not None
        entitydb.delete_entity(self._spatial_filter)
        self._spatial_filter = None

    def _discard_inverted_clip(self) -> None:
        if isinstance(self._spatial_filter, SpatialFilter):
            self._spatial_filter.discard_extension_dict()

    def cleanup(self):
        """Discard the extension dictionary of the base entity when empty."""
        self._insert.discard_empty_extension_dict()


def _rect_path(vertices: Iterable[Vec2]) -> Sequence[Vec2]:
    """Returns the path vertices for the smallest rectangular boundary around the given
    vertices.
    """
    return BoundingBox2d(vertices).rect_vertices()


def _get_inverted_clip_compare_vertices(
    bbox: BoundingBox2d, hole: Sequence[Vec2]
) -> Sequence[Vec2]:
    # AutoCAD does not accept this paths and from further tests it's clear that the 
    # geometry of the inverted clipping path is the problem not the DXF structure!
    from ezdxf.math.clipping import make_inverted_clipping_polygon

    assert (bbox.extmax is not None) and (bbox.extmin is not None)
    return make_inverted_clipping_polygon(inner_polygon=list(hole), outer_bounds=bbox)


def get_spatial_filter(entity: DXFEntity) -> SpatialFilter | None:
    """Returns the underlaying SPATIAL_FILTER entity if the given `entity` has a
    clipping path and returns ``None`` otherwise.
    """
    try:
        xdict = entity.get_extension_dict()
    except AttributeError:
        return None
    acad_filter = xdict.get(ACAD_FILTER)
    if not isinstance(acad_filter, Dictionary):
        return None
    acad_spatial_filter = acad_filter.get(SPATIAL)
    if isinstance(acad_spatial_filter, SpatialFilter):
        return acad_spatial_filter
    return None


def new_spatial_filter(entity: DXFEntity) -> SpatialFilter:
    """Creates the extension dict, the sub-dictionary ACAD_FILTER and the SPATIAL_FILTER
    entity if not exist.
    """
    doc = entity.doc
    if doc is None:
        raise const.DXFTypeError("Cannot add new clipping path to virtual entity.")
    try:
        xdict = entity.get_extension_dict()
    except AttributeError:
        xdict = entity.new_extension_dict()
    acad_filter_dict = xdict.dictionary.get_required_dict(ACAD_FILTER, hard_owned=True)
    spatial_filter = acad_filter_dict.get(SPATIAL)
    if not isinstance(spatial_filter, SpatialFilter):
        spatial_filter = doc.objects.add_dxf_object_with_reactor(
            "SPATIAL_FILTER", {"owner": acad_filter_dict.dxf.handle}
        )
        acad_filter_dict.add(SPATIAL, spatial_filter)
    assert isinstance(spatial_filter, SpatialFilter)
    return spatial_filter


def new_roundtrip_xrecord(spatial_filter: SpatialFilter) -> RoundtripXRecord:
    try:
        xdict = spatial_filter.get_extension_dict()
    except AttributeError:
        xdict = spatial_filter.new_extension_dict()
    xrec = xdict.get(ACAD_XREC_ROUNDTRIP)
    if xrec is None:
        xrec = xdict.add_xrecord(ACAD_XREC_ROUNDTRIP)
        xrec.set_reactors([xdict.handle])
    assert isinstance(xrec, XRecord)
    return RoundtripXRecord(xrec)


def get_roundtrip_xrecord(
    spatial_filter: SpatialFilter | None,
) -> RoundtripXRecord | None:
    if spatial_filter is None:
        return None
    try:
        xdict = spatial_filter.get_extension_dict()
    except AttributeError:
        return None
    xrecord = xdict.get(ACAD_XREC_ROUNDTRIP)
    if isinstance(xrecord, XRecord):
        return RoundtripXRecord(xrecord)
    return None


def get_roundtrip_vertices(
    xrec: RoundtripXRecord, section_name: str, m: Matrix44
) -> Sequence[Vec2]:
    tags = xrec.get_section(section_name)
    vertices = m.transform_vertices(Vec3(t.value) for t in tags)
    return Vec2.tuple(vertices)


def bbox_area(vertice: Sequence[Vec2]) -> float:
    bbox = BoundingBox2d(vertice)
    if bbox.has_data:
        size = bbox.size
        return size.x * size.y
    return 0.0



#  Copyright (c) 2020, Manfred Moitzi
#  License: MIT License
import pytest
from ezdxf.entities import factory
from ezdxf.render.forms import square, translate
from ezdxf.lldxf import const
from ezdxf.addons import geo

shapely_geometry = pytest.importorskip("shapely.geometry")


def test_shapely_geo_interface():
    point = shapely_geometry.shape(
        {
            "type": "Point",
            "coordinates": (0, 0),
        }
    )
    assert (point.x, point.y) == (0, 0)


def validate(geo_proxy: geo.GeoProxy) -> bool:
    return shapely_geometry.shape(geo_proxy).is_valid


def test_resolved_hatch_with_intersecting_holes():
    hatch = factory.new("HATCH")
    paths = hatch.paths
    paths.add_polyline_path(square(10), flags=const.BOUNDARY_PATH_EXTERNAL)
    paths.add_polyline_path(
        translate(square(3), (1, 1)), flags=const.BOUNDARY_PATH_DEFAULT
    )
    paths.add_polyline_path(
        translate(square(3), (2, 2)), flags=const.BOUNDARY_PATH_DEFAULT
    )
    p = geo.proxy(hatch)
    # Overlapping holes already resolved by fast_bbox_detection()
    polygon = shapely_geometry.shape(p)
    assert polygon.is_valid is True

    p.filter(validate)
    assert p.root["type"] == "Polygon"
    assert len(p.root["coordinates"]) == 2


def test_valid_hatch():
    hatch = factory.new("HATCH")
    paths = hatch.paths
    paths.add_polyline_path(square(10), flags=const.BOUNDARY_PATH_EXTERNAL)
    paths.add_polyline_path(
        translate(square(3), (1, 1)), flags=const.BOUNDARY_PATH_DEFAULT
    )
    paths.add_polyline_path(
        translate(square(3), (5, 1)), flags=const.BOUNDARY_PATH_DEFAULT
    )
    p = geo.proxy(hatch)
    polygon = shapely_geometry.shape(p)
    assert polygon.is_valid is True

    p.filter(validate)
    assert p.root != {}


def test_proxy_given_hatch_when_force_line_string_generates_valid_multi_line_string() -> None:
    hatch = factory.new("HATCH")
    paths = hatch.paths
    paths.add_polyline_path(square(10), flags=const.BOUNDARY_PATH_EXTERNAL)
    paths.add_polyline_path(
        translate(square(3), (1, 1)), flags=const.BOUNDARY_PATH_DEFAULT
    )
    p = geo.proxy(hatch, force_line_string=True)
    feature = p.__geo_interface__
    assert feature["type"] == "MultiLineString"

    # According to the __geo_interface__ specification:
    # https://gist.github.com/sgillies/2217756
    assert isinstance(feature["coordinates"][0][0], tuple)


if __name__ == "__main__":
    pytest.main([__file__])



#  Copyright (c) 2023, Manfred Moitzi
#  License: MIT License

import pytest
from ezdxf.fonts import lff


@pytest.fixture(scope="module")
def font() -> lff.LCFont:
    return lff.loads(EXAMPLE)


def test_load_font(font):
    assert font.name == "FontName"
    assert font.letter_spacing == 3.0
    assert font.word_spacing == 6.75
    assert len(font) == 5


def test_glyph_A(font):
    glyph = font.get(ord("A"))
    assert glyph.code == 65
    assert len(glyph.polylines) == 2

    p0, p1 = glyph.polylines
    assert len(p0) == 2
    assert len(p1) == 3

    assert p0[0][0] == pytest.approx(0.8333)
    assert p1[1][1] == pytest.approx(9)


def test_glyph_dollar(font):
    glyph = font.get(ord("$"))
    assert glyph.code == 36
    assert len(glyph.polylines) == 2
    assert len(glyph.polylines[1][1]) == 3
    bulge = glyph.polylines[1][1][2]
    assert bulge == pytest.approx(0.0935)


def test_composite_glyph(font):
    glyph = font.get(194)
    assert glyph.code == 194
    assert len(glyph.polylines) == 3

    p0, p1, p2 = glyph.polylines
    assert len(p0) == 2
    assert len(p1) == 3
    assert len(p2) == 3

    assert p0[0][0] == pytest.approx(0.8333)
    assert p1[1][1] == pytest.approx(9)
    assert p2[0][1] == pytest.approx(11.5)


def test_render_glyphs(font: lff.LCFont):
    p = font.get(112).to_path()
    assert p.has_curves is True
    assert p.start.isclose((0, -3))
    assert p.end.isclose((0, 0))


def test_scan_int_ex():
    assert lff.scan_int_ex("") == 0
    assert lff.scan_int_ex("[") == 0
    assert lff.scan_int_ex("[]") == 0
    assert lff.scan_int_ex("[a]") == 10
    assert lff.scan_int_ex("[a") == 10
    assert lff.scan_int_ex("[0000]") == 0
    assert lff.scan_int_ex("[000a]") == 10
    assert lff.scan_int_ex("[000A]") == 10
    assert lff.scan_int_ex("[000A") == 10
    assert lff.scan_int_ex("[#000A]") == 10
    assert lff.scan_int_ex("[#000A X") == 10
    assert lff.scan_int_ex("[# 0 0 0 A] X") == 10
    assert lff.scan_int_ex("[# 0. 0 0 A X") == 10
    assert lff.scan_int_ex("[# 0. 0 0 A a") != 10


EXAMPLE = """# Just a comment
# Name:              FontName
# LetterSpacing:     3
# WordSpacing:       6.75

[0024] $
2,0;2,9
0,1.5;2.39,1,A0.0935;3.235,3.81,A0.8;0.7685,5.19;1.6125,8,A-0.8;4,7.5,A-0.09

[0041] A
0.8333,2.5;5.1666,2.5
0,0;3,9;6,0

[0070] p
0,-3;0,6;2.5,6;4,4.5,A-0.4142;4,1.5;2.5,0,A-0.4142;0,0

[0078] x
0,0;4,6
0,6;4,0

[00c2] Â
C0041
1.5,11.5;3,13;4.5,11.5

"""
if __name__ == "__main__":
    pytest.main([__file__])



# Copyright (c) 2019 Manfred Moitzi
# License: MIT License
# created 2019-03-05
import pytest

from ezdxf.entities.xline import Ray
from ezdxf.lldxf.tagwriter import TagCollector, basic_tags_from_text

RAY = """0
RAY
5
0
330
0
100
AcDbEntity
8
0
100
AcDbRay
10
0.0
20
0.0
30
0.0
11
1.0
21
0.0
31
0.0
"""


@pytest.fixture
def entity():
    return Ray.from_text(RAY)


def test_registered():
    from ezdxf.entities.factory import ENTITY_CLASSES

    assert "RAY" in ENTITY_CLASSES


def test_default_init():
    entity = Ray()
    assert entity.dxftype() == "RAY"
    assert entity.dxf.handle is None
    assert entity.dxf.owner is None


def test_default_new():
    entity = Ray.new(
        handle="ABBA",
        owner="0",
        dxfattribs={
            "color": 7,
            "start": (1, 2, 3),
            "unit_vector": (4, 5, 6),
        },
    )
    assert entity.dxf.layer == "0"
    assert entity.dxf.color == 7
    assert entity.dxf.start == (1, 2, 3)
    assert entity.dxf.unit_vector == (4, 5, 6)


def test_load_from_text(entity):
    assert entity.dxf.layer == "0"
    assert entity.dxf.color == 256, "default color is 256 (by layer)"
    assert entity.dxf.start == (0, 0, 0)
    assert entity.dxf.unit_vector == (1, 0, 0)


def test_write_dxf():
    entity = Ray.from_text(RAY)
    result = TagCollector.dxftags(entity)
    expected = basic_tags_from_text(RAY)
    assert result == expected



#  Copyright (c) 2023, Manfred Moitzi
#  License: MIT License
from __future__ import annotations
from typing import Iterable, Optional, Iterator, Sequence
from typing_extensions import Self, TypeAlias
import abc

import numpy as np
import numpy.typing as npt

from ezdxf.math import (
    Matrix44,
    UVec,
    Vec2,
    Vec3,
    has_clockwise_orientation,
    Bezier3P,
    Bezier4P,
    BoundingBox2d,
    BoundingBox,
)
from ezdxf.path import (
    Path,
    Command,
    PathElement,
    LineTo,
    MoveTo,
    Curve3To,
    Curve4To,
    nesting,
)

try:
    from ezdxf.acc import np_support
except ImportError:
    np_support = None

__all__ = [
    "NumpyPath2d",
    "NumpyPoints2d",
    "NumpyPoints3d",
    "NumpyShapesException",
    "EmptyShapeError",
    "to_qpainter_path",
    "to_matplotlib_path",
    "single_paths",
    "orient_paths",
]

# comparing Command.<attrib> to ints is very slow
CMD_MOVE_TO = int(Command.MOVE_TO)
CMD_LINE_TO = int(Command.LINE_TO)
CMD_CURVE3_TO = int(Command.CURVE3_TO)
CMD_CURVE4_TO = int(Command.CURVE4_TO)


class NumpyShapesException(Exception):
    pass


class EmptyShapeError(NumpyShapesException):
    pass


CommandNumpyType: TypeAlias = np.int8
VertexNumpyType: TypeAlias = np.float64
EMPTY_SHAPE = np.array([], dtype=VertexNumpyType)
NO_COMMANDS = np.array([], dtype=CommandNumpyType)


class NumpyShape2d(abc.ABC):
    """This is an optimization to store many 2D paths and polylines in a compact way
    without sacrificing basic functions like transformation and bounding box calculation.
    """

    _vertices: npt.NDArray[VertexNumpyType] = EMPTY_SHAPE

    def extents(self) -> tuple[Vec2, Vec2]:
        """Returns the extents of the bounding box as tuple (extmin, extmax)."""
        v = self._vertices
        if len(v) > 0:
            return Vec2(v.min(0)), Vec2(v.max(0))
        else:
            raise EmptyShapeError("empty shape has no extends")

    @abc.abstractmethod
    def clone(self) -> Self:
        ...

    def np_vertices(self) -> npt.NDArray[VertexNumpyType]:
        return self._vertices

    def transform_inplace(self, m: Matrix44) -> None:
        """Transforms the vertices of the shape inplace."""
        v = self._vertices
        if len(v) == 0:
            return
        m.transform_array_inplace(v, 2)

    def vertices(self) -> list[Vec2]:
        """Returns the shape vertices as list of :class:`Vec2` 
        e.g. [Vec2(1, 2), Vec2(3, 4), ...] 
        """
        return [Vec2(v) for v in self._vertices]

    def to_tuples(self) -> list[tuple[float, float]]:
        """Returns the shape vertices as list of 2-tuples 
        e.g. [(1, 2), (3, 4), ...]
        """
        return [tuple(v) for v in self._vertices]
    
    def to_list(self) -> list[list[float]]:
        """Returns the shape vertices as list of lists 
        e.g. [[1, 2], [3, 4], ...]
        """
        return self._vertices.tolist()  # type: ignore
    
    def bbox(self) -> BoundingBox2d:
        """Returns the bounding box of all vertices."""
        return BoundingBox2d(self.extents())


class NumpyPoints2d(NumpyShape2d):
    """Represents an array of 2D points stored as a ndarray."""

    def __init__(self, points: Optional[Iterable[Vec2 | Vec3]]) -> None:
        if points:
            self._vertices = np.array(
                [(v.x, v.y) for v in points], dtype=VertexNumpyType
            )

    def clone(self) -> Self:
        clone = self.__class__(None)
        clone._vertices = self._vertices.copy()
        return clone

    __copy__ = clone

    def __len__(self) -> int:
        return len(self._vertices)


class NumpyPath2d(NumpyShape2d):
    """Represents a 2D path, the path control vertices and commands are stored as ndarray.

    This class cannot build paths from scratch and is therefore not a drop-in replacement
    for the :class:`ezdxf.path.Path` class. Operations like transform and reverse are
    done inplace to utilize the `numpy` capabilities. This behavior is different from the
    :class:`ezdxf.path.Path` class!!!

    Construct new paths by the :class:`Path` class and convert them to
    :class:`NumpyPath2d` instances::

        path = Path((0, 0))
        path.line_to((50, 70))
        ...
        path2d = NumpyPath2d(path)

    """

    _commands: npt.NDArray[CommandNumpyType]

    def __init__(self, path: Optional[Path]) -> None:
        if path is None:
            self._vertices = EMPTY_SHAPE
            self._commands = NO_COMMANDS
            return
        # (v.x, v.y) is 4x faster than Vec2(v), see profiling/numpy_array_setup.py
        vertices = [(v.x, v.y) for v in path.control_vertices()]
        if len(vertices) == 0:
            try:  # control_vertices() does not return the start point of empty paths
                vertices = [Vec2(path.start)]
            except IndexError:
                vertices = []
        self._vertices = np.array(vertices, dtype=VertexNumpyType)
        self._commands = np.array(path.command_codes(), dtype=CommandNumpyType)

    def __len__(self) -> int:
        return len(self._commands)

    @property
    def start(self) -> Vec2:
        """Returns the start point as :class:`~ezdxf.math.Vec2` instance."""
        return Vec2(self._vertices[0])

    @property
    def end(self) -> Vec2:
        """Returns the end point as :class:`~ezdxf.math.Vec2` instance."""
        return Vec2(self._vertices[-1])

    def control_vertices(self) -> list[Vec2]:
        return [Vec2(v) for v in self._vertices]

    def clone(self) -> Self:
        clone = self.__class__(None)
        clone._commands = self._commands.copy()
        clone._vertices = self._vertices.copy()
        return clone

    __copy__ = clone

    def command_codes(self) -> list[int]:
        """Internal API."""
        return list(self._commands)

    def commands(self) -> Iterator[PathElement]:
        vertices = self.vertices()
        index = 1
        for cmd in self._commands:
            if cmd == CMD_LINE_TO:
                yield LineTo(vertices[index])
                index += 1
            elif cmd == CMD_CURVE3_TO:
                yield Curve3To(vertices[index + 1], vertices[index])
                index += 2
            elif cmd == CMD_CURVE4_TO:
                yield Curve4To(
                    vertices[index + 2], vertices[index], vertices[index + 1]
                )
                index += 3
            elif cmd == CMD_MOVE_TO:
                yield MoveTo(vertices[index])
                index += 1

    def to_path(self) -> Path:
        """Returns a new :class:`ezdxf.path.Path` instance."""
        vertices = [Vec3(v) for v in self._vertices]
        commands = [Command(c) for c in self._commands]
        return Path.from_vertices_and_commands(vertices, commands)

    @classmethod
    def from_vertices(
        cls, vertices: Iterable[Vec2 | Vec3], close: bool = False
    ) -> Self:
        new_path = cls(None)
        vertices = list(vertices)
        if len(vertices) == 0:
            return new_path
        if close and not vertices[0].isclose(vertices[-1]):
            vertices.append(vertices[0])
        # (v.x, v.y) is 4x faster than Vec2(v), see profiling/numpy_array_setup.py
        points = [(v.x, v.y) for v in vertices]
        new_path._vertices = np.array(points, dtype=VertexNumpyType)
        new_path._commands = np.full(
            len(points) - 1, fill_value=CMD_LINE_TO, dtype=CommandNumpyType
        )
        return new_path

    @property
    def has_sub_paths(self) -> bool:
        """Returns ``True`` if the path is a :term:`Multi-Path` object that
        contains multiple sub-paths.

        """
        return CMD_MOVE_TO in self._commands

    @property
    def is_closed(self) -> bool:
        """Returns ``True`` if the start point is close to the end point."""
        if len(self._vertices) > 1:
            return self.start.isclose(self.end)
        return False

    @property
    def has_lines(self) -> bool:
        """Returns ``True`` if the path has any line segments."""
        return CMD_LINE_TO in self._commands

    @property
    def has_curves(self) -> bool:
        """Returns ``True`` if the path has any curve segments."""
        return CMD_CURVE3_TO in self._commands or CMD_CURVE4_TO in self._commands

    def sub_paths(self) -> list[Self]:
        """Yield all sub-paths as :term:`Single-Path` objects.

        It's safe to call :meth:`sub_paths` on any path-type:
        :term:`Single-Path`, :term:`Multi-Path` and :term:`Empty-Path`.

        """

        def append_sub_path() -> None:
            s: Self = self.__class__(None)
            s._vertices = vertices[vtx_start_index : vtx_index + 1]  # .copy() ?
            s._commands = commands[cmd_start_index:cmd_index]  # .copy() ?
            sub_paths.append(s)

        commands = self._commands
        if len(commands) == 0:
            return []
        if CMD_MOVE_TO not in commands:
            return [self]

        sub_paths: list[Self] = []
        vertices = self._vertices
        vtx_start_index = 0
        vtx_index = 0
        cmd_start_index = 0
        cmd_index = 0
        for cmd in commands:
            if cmd == CMD_LINE_TO:
                vtx_index += 1
            elif cmd == CMD_CURVE3_TO:
                vtx_index += 2
            elif cmd == CMD_CURVE4_TO:
                vtx_index += 3
            elif cmd == CMD_MOVE_TO:
                append_sub_path()
                # MOVE_TO target vertex is the start vertex of the following path.
                vtx_index += 1
                vtx_start_index = vtx_index
                cmd_start_index = cmd_index + 1
            cmd_index += 1

        if commands[-1] != CMD_MOVE_TO:
            append_sub_path()
        return sub_paths

    def has_clockwise_orientation(self) -> bool:
        """Returns ``True`` if 2D path has clockwise orientation.

        Raises:
            TypeError: can't detect orientation of a :term:`Multi-Path` object

        """
        if self.has_sub_paths:
            raise TypeError("can't detect orientation of a multi-path object")
        if np_support is None:
            return has_clockwise_orientation(self.vertices())
        else:
            return np_support.has_clockwise_orientation(self._vertices)

    def reverse(self) -> Self:
        """Reverse path orientation inplace."""
        commands = self._commands
        if not len(self._commands):
            return self
        if commands[-1] == CMD_MOVE_TO:
            # The last move_to will become the first move_to.
            # A move_to as first command just moves the start point and can be
            # removed!
            # There are never two consecutive MOVE_TO commands in a Path!
            self._commands = np.flip(commands[:-1]).copy()
            self._vertices = np.flip(self._vertices[:-1, ...], axis=0).copy()
        else:
            self._commands = np.flip(commands).copy()
            self._vertices = np.flip(self._vertices, axis=0).copy()
        return self

    def clockwise(self) -> Self:
        """Apply clockwise orientation inplace.

        Raises:
            TypeError: can't detect orientation of a :term:`Multi-Path` object

        """
        if not self.has_clockwise_orientation():
            self.reverse()
        return self

    def counter_clockwise(self) -> Self:
        """Apply counter-clockwise orientation inplace.

        Raises:
            TypeError: can't detect orientation of a :term:`Multi-Path` object

        """

        if self.has_clockwise_orientation():
            self.reverse()
        return self

    def flattening(self, distance: float, segments: int = 4) -> Iterator[Vec2]:
        """Flatten path to vertices as :class:`Vec2` instances."""
        if not len(self._commands):
            return

        vertices = self.vertices()
        start = vertices[0]
        yield start
        index = 1
        for cmd in self._commands:
            if cmd == CMD_LINE_TO or cmd == CMD_MOVE_TO:
                end_location = vertices[index]
                index += 1
                yield end_location
            elif cmd == CMD_CURVE3_TO:
                ctrl, end_location = vertices[index : index + 2]
                index += 2
                pts = Vec2.generate(
                    Bezier3P((start, ctrl, end_location)).flattening(distance, segments)
                )
                next(pts)  # skip first vertex
                yield from pts
            elif cmd == CMD_CURVE4_TO:
                ctrl1, ctrl2, end_location = vertices[index : index + 3]
                index += 3
                pts = Vec2.generate(
                    Bezier4P((start, ctrl1, ctrl2, end_location)).flattening(
                        distance, segments
                    )
                )
                next(pts)  # skip first vertex
                yield from pts
            else:
                raise ValueError(f"Invalid command: {cmd}")
            start = end_location

    # Appending single commands (line_to, move_to, curve3_to, curve4_to) is not
    # efficient, because numpy arrays do not grow dynamically, they are reallocated for
    # every single command!
    # Construct paths as ezdxf.path.Path and convert them to NumpyPath2d.
    # Concatenation of NumpyPath2d objects is faster than extending Path objects

    def extend(self, paths: Sequence[NumpyPath2d]) -> None:
        """Extend an existing path by appending additional paths. The paths are
        connected by MOVE_TO commands if the end- and start point of sequential paths
        are not coincident (multi-path).
        """
        if not len(paths):
            return
        if not len(self._commands):
            first = paths[0]
            paths = paths[1:]
        else:
            first = self

        vertices: list[np.ndarray] = [first._vertices]
        commands: list[np.ndarray] = [first._commands]
        end: Vec2 = first.end

        for next_path in paths:
            if len(next_path._commands) == 0:
                continue
            if not end.isclose(next_path.start):
                commands.append(np.array((CMD_MOVE_TO,), dtype=CommandNumpyType))
                vertices.append(next_path._vertices)
            else:
                vertices.append(next_path._vertices[1:])
            end = next_path.end
            commands.append(next_path._commands)
        self._vertices = np.concatenate(vertices, axis=0)
        self._commands = np.concatenate(commands)

    @staticmethod
    def concatenate(paths: Sequence[NumpyPath2d]) -> NumpyPath2d:
        """Returns a new path of concatenated paths. The paths are connected by
        MOVE_TO commands if the end- and start point of sequential paths are not
        coincident (multi-path).
        """

        if not paths:
            return NumpyPath2d(None)
        first = paths[0].clone()
        first.extend(paths[1:])
        return first


def to_qpainter_path(paths: Iterable[NumpyPath2d]):
    """Convert the given `paths` into a single :class:`QPainterPath`."""
    from ezdxf.addons.xqt import QPainterPath, QPointF

    paths = list(paths)
    if len(paths) == 0:
        raise ValueError("one or more paths required")

    qpath = QPainterPath()
    for path in paths:
        points = [QPointF(v.x, v.y) for v in path.vertices()]
        qpath.moveTo(points[0])
        index = 1
        for cmd in path.command_codes():
            # using Command.<attr> slows down this function by a factor of 4!!!
            if cmd == CMD_LINE_TO:
                qpath.lineTo(points[index])
                index += 1
            elif cmd == CMD_CURVE3_TO:
                qpath.quadTo(points[index], points[index + 1])
                index += 2
            elif cmd == CMD_CURVE4_TO:
                qpath.cubicTo(points[index], points[index + 1], points[index + 2])
                index += 3
            elif cmd == CMD_MOVE_TO:
                qpath.moveTo(points[index])
                index += 1
    return qpath


MPL_MOVETO = 1
MPL_LINETO = 2
MPL_CURVE3 = 3
MPL_CURVE4 = 4

MPL_CODES = [
    (0,),  # dummy
    (MPL_LINETO,),
    (MPL_CURVE3, MPL_CURVE3),
    (MPL_CURVE4, MPL_CURVE4, MPL_CURVE4),
    (MPL_MOVETO,),
]


def to_matplotlib_path(paths: Iterable[NumpyPath2d], *, detect_holes=False):
    """Convert the given `paths` into a single :class:`matplotlib.path.Path`.

    Matplotlib requires counter-clockwise oriented outside paths and clockwise oriented
    holes. Set the `detect_holes` argument to ``True`` if this path orientation is not
    yet satisfied.
    """
    from matplotlib.path import Path

    paths = list(paths)
    if len(paths) == 0:
        raise ValueError("one or more paths required")

    if detect_holes:
        # path orientation for holes is important, see #939
        paths = orient_paths(paths)

    vertices: list[np.ndarray] = []
    codes: list[int] = []
    for path in paths:
        vertices.append(path.np_vertices())
        codes.append(MPL_MOVETO)
        for cmd in path.command_codes():
            codes.extend(MPL_CODES[cmd])
    points = np.concatenate(vertices)
    try:
        return Path(points, codes)
    except Exception as e:
        raise ValueError(f"matplotlib.path.Path({str(points)}, {str(codes)}): {str(e)}")


def single_paths(paths: Iterable[NumpyPath2d]) -> list[NumpyPath2d]:
    single_paths_: list[NumpyPath2d] = []
    for p in paths:
        sub_paths = p.sub_paths()
        if sub_paths:
            single_paths_.extend(sub_paths)
    return single_paths_


def orient_paths(paths: list[NumpyPath2d]) -> list[NumpyPath2d]:
    """Returns a new list of paths, with outer paths oriented counter-clockwise and
    holes oriented clockwise.
    """
    sub_paths: list[NumpyPath2d] = single_paths(paths)
    if len(sub_paths) < 2:
        return paths

    polygons = nesting.make_polygon_structure(sub_paths)
    outer_paths: list[NumpyPath2d]
    holes: list[NumpyPath2d]
    outer_paths, holes = nesting.winding_deconstruction(polygons)

    path: NumpyPath2d
    for path in outer_paths:
        path.counter_clockwise()
    for path in holes:
        path.clockwise()
    return outer_paths + holes


class NumpyShape3d(abc.ABC):
    """This is an optimization to store many 3D paths and polylines in a compact way
    without sacrificing basic functions like transformation and bounding box calculation.
    """

    _vertices: npt.NDArray[VertexNumpyType] = EMPTY_SHAPE

    def extents(self) -> tuple[Vec3, Vec3]:
        """Returns the extents of the bounding box as tuple (extmin, extmax)."""
        v = self._vertices
        if len(v) > 0:
            return Vec3(v.min(0)), Vec3(v.max(0))
        else:
            raise EmptyShapeError("empty shape has no extends")

    @abc.abstractmethod
    def clone(self) -> Self:
        ...

    def np_vertices(self) -> npt.NDArray[VertexNumpyType]:
        return self._vertices

    def transform_inplace(self, m: Matrix44) -> None:
        """Transforms the vertices of the shape inplace."""
        v = self._vertices
        if len(v) == 0:
            return
        m.transform_array_inplace(v, 3)

    def vertices(self) -> list[Vec3]:
        """Returns the shape vertices as list of :class:`Vec3`."""
        return [Vec3(v) for v in self._vertices]

    def bbox(self) -> BoundingBox:
        """Returns the bounding box of all vertices."""
        return BoundingBox(self.extents())


class NumpyPoints3d(NumpyShape3d):
    """Represents an array of 3D points stored as a ndarray."""

    def __init__(self, points: Optional[Iterable[UVec]]) -> None:
        if points:
            self._vertices = np.array(
                [Vec3(v).xyz for v in points], dtype=VertexNumpyType
            )

    def clone(self) -> Self:
        clone = self.__class__(None)
        clone._vertices = self._vertices.copy()
        return clone

    __copy__ = clone

    def __len__(self) -> int:
        return len(self._vertices)



# Copyright (c) 2011-2023, Manfred Moitzi
# License: MIT License
from __future__ import annotations
from typing import (
    Iterable,
    Iterator,
    Optional,
    TYPE_CHECKING,
    Union,
    cast,
)
import logging

from ezdxf.entities.dictionary import Dictionary
from ezdxf.entities import factory, is_dxf_object
from ezdxf.lldxf import const, validator
from ezdxf.entitydb import EntitySpace, EntityDB
from ezdxf.query import EntityQuery
from ezdxf.tools.handle import UnderlayKeyGenerator
from ezdxf.audit import Auditor, AuditError

if TYPE_CHECKING:
    from ezdxf.document import Drawing
    from ezdxf.lldxf.tagwriter import AbstractTagWriter
    from ezdxf.entities import (
        GeoData,
        DictionaryVar,
        DXFTagStorage,
        DXFObject,
        ImageDef,
        UnderlayDefinition,
        DictionaryWithDefault,
        XRecord,
        Placeholder,
    )
    from ezdxf.entities.image import ImageDefReactor

logger = logging.getLogger("ezdxf")


class ObjectsSection:
    def __init__(self, doc: Drawing, entities: Optional[Iterable[DXFObject]] = None):
        self.doc = doc
        self.underlay_key_generator = UnderlayKeyGenerator()
        self._entity_space = EntitySpace()
        if entities is not None:
            self._build(iter(entities))

    @property
    def entitydb(self) -> EntityDB:
        """Returns drawing entity database. (internal API)"""
        return self.doc.entitydb

    def get_entity_space(self) -> EntitySpace:
        """Returns entity space. (internal API)"""
        return self._entity_space

    def next_underlay_key(self, checkfunc=lambda k: True) -> str:
        while True:
            key = self.underlay_key_generator.next()
            if checkfunc(key):
                return key

    def _build(self, entities: Iterator[DXFObject]) -> None:
        section_head = cast("DXFTagStorage", next(entities))

        if section_head.dxftype() != "SECTION" or section_head.base_class[1] != (
            2,
            "OBJECTS",
        ):
            raise const.DXFStructureError(
                "Critical structure error in the OBJECTS section."
            )

        for entity in entities:
            # No check for valid entities here:
            # Use the audit or the recover module to fix invalid DXF files!
            self._entity_space.add(entity)

    def export_dxf(self, tagwriter: AbstractTagWriter) -> None:
        """Export DXF entity by `tagwriter`. (internal API)"""
        tagwriter.write_str("  0\nSECTION\n  2\nOBJECTS\n")
        self._entity_space.export_dxf(tagwriter)
        tagwriter.write_tag2(0, "ENDSEC")

    def new_entity(self, _type: str, dxfattribs: dict) -> DXFObject:
        """Create new DXF object, add it to the entity database and to the
        entity space.

        Args:
             _type: DXF type like `DICTIONARY`
             dxfattribs: DXF attributes as dict

        (internal API)
        """
        dxf_entity = factory.create_db_entry(_type, dxfattribs, self.doc)
        self._entity_space.add(dxf_entity)
        return dxf_entity  # type: ignore

    def delete_entity(self, entity: DXFObject) -> None:
        """Remove `entity` from entity space and destroy object. (internal API)"""
        self._entity_space.remove(entity)
        self.entitydb.delete_entity(entity)

    def delete_all_entities(self) -> None:
        """Delete all DXF objects. (internal API)"""
        db = self.entitydb
        for entity in self._entity_space:
            db.delete_entity(entity)
        self._entity_space.clear()

    def setup_rootdict(self) -> Dictionary:
        """Create a root dictionary. Has to be the first object in the objects
        section. (internal API)"""
        if len(self):
            raise const.DXFStructureError(
                "Can not create root dictionary in none empty objects section."
            )
        logger.debug("Creating ROOT dictionary.")
        # root directory has no owner
        return self.add_dictionary(owner="0", hard_owned=False)

    def setup_object_management_tables(self, rootdict: Dictionary) -> None:
        """Setup required management tables. (internal API)"""

        def setup_plot_style_name_table():
            plot_style_name_dict = self.add_dictionary_with_default(
                owner=rootdict.dxf.handle,
                hard_owned=False,
            )
            placeholder = self.add_placeholder(owner=plot_style_name_dict.dxf.handle)
            plot_style_name_dict.set_default(placeholder)
            plot_style_name_dict["Normal"] = placeholder
            rootdict["ACAD_PLOTSTYLENAME"] = plot_style_name_dict

        def restore_table_handle(table_name: str, handle: str) -> None:
            # The original object table does not exist, but the handle is maybe
            # used by some DXF entities. Try to restore the original table
            # handle.
            new_table = rootdict.get(table_name)
            if isinstance(new_table, Dictionary) and self.doc.entitydb.reset_handle(
                new_table, handle
            ):
                logger.debug(f"reset handle of table {table_name} to #{handle}")

        # check required object tables:
        for name in _OBJECT_TABLE_NAMES:
            table: Union[Dictionary, str, None] = rootdict.get(name)  # type: ignore
            # Dictionary: existing table object
            # str: handle of not existing table object
            # None: no entry in the rootdict for "name"
            if isinstance(table, Dictionary):
                continue  # skip existing tables

            logger.info(f"creating {name} dictionary")
            if name == "ACAD_PLOTSTYLENAME":
                setup_plot_style_name_table()
            else:
                rootdict.add_new_dict(name, hard_owned=False)

            if isinstance(table, str) and validator.is_handle(table):
                restore_table_handle(name, handle=table)

    def add_object(self, entity: DXFObject) -> None:
        """Add `entity` to OBJECTS section. (internal API)"""
        if is_dxf_object(entity):
            self._entity_space.add(entity)
        else:
            raise const.DXFTypeError(
                f"invalid DXF type {entity.dxftype()} for OBJECTS section"
            )

    def add_dxf_object_with_reactor(self, dxftype: str, dxfattribs) -> DXFObject:
        """Add DXF object with reactor. (internal API)"""
        dxfobject = self.new_entity(dxftype, dxfattribs)
        dxfobject.set_reactors([dxfattribs["owner"]])
        return dxfobject

    def purge(self):
        self._entity_space.purge()

    # start of public interface

    @property
    def rootdict(self) -> Dictionary:
        """Returns the root DICTIONARY, or as AutoCAD calls it:
        the named DICTIONARY.
        """
        if len(self):
            return self._entity_space[0]  # type: ignore
        else:
            return self.setup_rootdict()

    def __len__(self) -> int:
        """Returns the count of all DXF objects in the OBJECTS section."""
        return len(self._entity_space)

    def __iter__(self) -> Iterator[DXFObject]:
        """Returns an iterator of all DXF objects in the OBJECTS section."""
        return iter(self._entity_space)  # type: ignore

    def __getitem__(self, index) -> DXFObject:
        """Get entity at `index`.

        The underlying data structure for storing DXF objects is organized like
        a standard Python list, therefore `index` can be any valid list indexing
        or slicing term, like a single index ``objects[-1]`` to get the last
        entity, or an index slice ``objects[:10]`` to get the first 10 or fewer
        objects as ``list[DXFObject]``.

        """
        return self._entity_space[index]  # type: ignore

    def __contains__(self, entity):
        """Returns ``True`` if `entity` stored in OBJECTS section.

        Args:
             entity: :class:`DXFObject` or handle as hex string

        """
        if isinstance(entity, str):
            try:
                entity = self.entitydb[entity]
            except KeyError:
                return False
        return entity in self._entity_space

    def query(self, query: str = "*") -> EntityQuery:
        """Get all DXF objects matching the :ref:`entity query string`."""
        return EntityQuery(iter(self), query)

    def audit(self, auditor: Auditor) -> None:
        """Audit and repair OBJECTS section.

        .. important::

            Do not delete entities while auditing process, because this
            would alter the entity database while iterating, instead use::

                auditor.trash(entity)

            to delete invalid entities after auditing automatically.

        """
        assert self.doc is auditor.doc, "Auditor for different DXF document."
        for entity in self._entity_space:
            if not is_dxf_object(entity):
                auditor.fixed_error(
                    code=AuditError.REMOVED_INVALID_DXF_OBJECT,
                    message=f"Removed invalid DXF entity {str(entity)} "
                    f"from OBJECTS section.",
                )
                auditor.trash(entity)
        self.reorg(auditor)
        self._entity_space.audit(auditor)

    def add_dictionary(self, owner: str = "0", hard_owned: bool = True) -> Dictionary:
        """Add new :class:`~ezdxf.entities.Dictionary` object.

        Args:
            owner: handle to owner as hex string.
            hard_owned: ``True`` to treat entries as hard owned.

        """
        entity = self.new_entity(
            "DICTIONARY",
            dxfattribs={
                "owner": owner,
                "hard_owned": hard_owned,
            },
        )
        return cast(Dictionary, entity)

    def add_dictionary_with_default(
        self, owner="0", default="0", hard_owned: bool = True
    ) -> DictionaryWithDefault:
        """Add new :class:`~ezdxf.entities.DictionaryWithDefault` object.

        Args:
            owner: handle to owner as hex string.
            default: handle to default entry.
            hard_owned: ``True`` to treat entries as hard owned.

        """
        entity = self.new_entity(
            "ACDBDICTIONARYWDFLT",
            dxfattribs={
                "owner": owner,
                "default": default,
                "hard_owned": hard_owned,
            },
        )
        return cast("DictionaryWithDefault", entity)

    def add_dictionary_var(self, owner: str = "0", value: str = "") -> DictionaryVar:
        """Add a new :class:`~ezdxf.entities.DictionaryVar` object.

        Args:
            owner: handle to owner as hex string.
            value: value as string

        """
        return self.new_entity(  # type: ignore
            "DICTIONARYVAR", dxfattribs={"owner": owner, "value": value}
        )

    def add_xrecord(self, owner: str = "0") -> XRecord:
        """Add a new :class:`~ezdxf.entities.XRecord` object.

        Args:
            owner: handle to owner as hex string.

        """
        return self.new_entity("XRECORD", dxfattribs={"owner": owner})  # type: ignore

    def add_placeholder(self, owner: str = "0") -> Placeholder:
        """Add a new :class:`~ezdxf.entities.Placeholder` object.

        Args:
            owner: handle to owner as hex string.

        """
        return self.new_entity(  # type: ignore
            "ACDBPLACEHOLDER", dxfattribs={"owner": owner}
        )

    # end of public interface

    def set_raster_variables(
        self, frame: int = 0, quality: int = 1, units: str = "m"
    ) -> None:
        """Set raster variables.

        Args:
            frame: ``0`` = do not show image frame; ``1`` = show image frame
            quality: ``0`` = draft; ``1`` = high
            units: units for inserting images. This defines the real world unit for one
                drawing unit for the purpose of inserting and scaling images with an
                associated resolution.

                   ===== ===========================
                   mm    Millimeter
                   cm    Centimeter
                   m     Meter (ezdxf default)
                   km    Kilometer
                   in    Inch
                   ft    Foot
                   yd    Yard
                   mi    Mile
                   none  None
                   ===== ===========================

        (internal API), public interface :meth:`~ezdxf.drawing.Drawing.set_raster_variables`

        """
        units_: int = const.RASTER_UNITS.get(units, 0)
        try:
            raster_vars = self.rootdict["ACAD_IMAGE_VARS"]
        except const.DXFKeyError:
            raster_vars = self.add_dxf_object_with_reactor(
                "RASTERVARIABLES",
                dxfattribs={
                    "owner": self.rootdict.dxf.handle,
                    "frame": frame,
                    "quality": quality,
                    "units": units_,
                },
            )
            self.rootdict["ACAD_IMAGE_VARS"] = raster_vars
        else:
            raster_vars.dxf.frame = frame
            raster_vars.dxf.quality = quality
            raster_vars.dxf.units = units_

    def get_raster_variables(self) -> tuple[int, int, str]:
        try:
            raster_vars = self.rootdict["ACAD_IMAGE_VARS"]
        except const.DXFKeyError:
            return 0, 1, "none"
        return (
            raster_vars.dxf.frame,
            raster_vars.dxf.quality,
            const.REVERSE_RASTER_UNITS.get(raster_vars.dxf.units, "none"),
        )

    def set_wipeout_variables(self, frame: int = 0) -> None:
        """Set wipeout variables.

        Args:
            frame: ``0`` = do not show image frame; ``1`` = show image frame

        (internal API)
        """
        try:
            wipeout_vars = self.rootdict["ACAD_WIPEOUT_VARS"]
        except const.DXFKeyError:
            wipeout_vars = self.add_dxf_object_with_reactor(
                "WIPEOUTVARIABLES",
                dxfattribs={
                    "owner": self.rootdict.dxf.handle,
                    "frame": int(frame),
                },
            )
            self.rootdict["ACAD_WIPEOUT_VARS"] = wipeout_vars
        else:
            wipeout_vars.dxf.frame = int(frame)

    def get_wipeout_frame_setting(self) -> int:
        try:
            wipeout_vars = self.rootdict["ACAD_WIPEOUT_VARS"]
        except const.DXFKeyError:
            return 0
        return wipeout_vars.dxf.frame

    def add_image_def(
        self,
        filename: str,
        size_in_pixel: tuple[int, int],
        name: Optional[str] = None,
    ) -> ImageDef:
        """Add an image definition to the objects section.

        Add an :class:`~ezdxf.entities.image.ImageDef` entity to the drawing
        (objects section). `filename` is the image file name as relative or
        absolute path and `size_in_pixel` is the image size in pixel as (x, y)
        tuple. To avoid dependencies to external packages, `ezdxf` can not
        determine the image size by itself. Returns a :class:`~ezdxf.entities.image.ImageDef`
        entity which is needed to create an image reference. `name` is the
        internal image name, if set to ``None``, name is auto-generated.

        Absolute image paths works best for AutoCAD but not really good, you
        have to update external references manually in AutoCAD, which is not
        possible in TrueView. If the drawing units differ from 1 meter, you also
        have to use: :meth:`set_raster_variables`.

        Args:
            filename: image file name (absolute path works best for AutoCAD)
            size_in_pixel: image size in pixel as (x, y) tuple
            name: image name for internal use, None for using filename as name
                (best for AutoCAD)

        """
        if name is None:
            name = filename
        image_dict = self.rootdict.get_required_dict("ACAD_IMAGE_DICT")
        image_def = self.add_dxf_object_with_reactor(
            "IMAGEDEF",
            dxfattribs={
                "owner": image_dict.dxf.handle,
                "filename": filename,
                "image_size": size_in_pixel,
            },
        )
        image_dict[name] = image_def
        return cast("ImageDef", image_def)

    def add_image_def_reactor(self, image_handle: str) -> ImageDefReactor:
        """Add required IMAGEDEF_REACTOR object for IMAGEDEF object.

        (internal API)
        """
        image_def_reactor = self.new_entity(
            "IMAGEDEF_REACTOR",
            dxfattribs={
                "owner": image_handle,
                "image_handle": image_handle,
            },
        )
        return cast("ImageDefReactor", image_def_reactor)

    def add_underlay_def(
        self, filename: str, fmt: str = "pdf", name: Optional[str] = None
    ) -> UnderlayDefinition:
        """Add an :class:`~ezdxf.entities.underlay.UnderlayDefinition` entity
        to the drawing (OBJECTS section). `filename` is the underlay file name
        as relative or absolute path and `fmt` as string (pdf, dwf, dgn).
        The underlay definition is required to create an underlay reference.

        Args:
            filename: underlay file name
            fmt: file format as string ``'pdf'|'dwf'|'dgn'``
            name: pdf format = page number to display; dgn format = ``'default'``; dwf: ????

        """
        fmt = fmt.upper()
        if fmt in ("PDF", "DWF", "DGN"):
            underlay_dict_name = f"ACAD_{fmt}DEFINITIONS"
            underlay_def_entity = f"{fmt}DEFINITION"
        else:
            raise const.DXFValueError(f"Unsupported file format: '{fmt}'")

        if name is None:
            if fmt == "PDF":
                name = "1"  # Display first page by default
            elif fmt == "DGN":
                name = "default"
            else:
                name = "Model"  # Display model space for DWF ???

        underlay_dict = self.rootdict.get_required_dict(underlay_dict_name)
        underlay_def = self.new_entity(
            underlay_def_entity,
            dxfattribs={
                "owner": underlay_dict.dxf.handle,
                "filename": filename,
                "name": name,
            },
        )

        # auto-generated underlay key
        key = self.next_underlay_key(lambda k: k not in underlay_dict)
        underlay_dict[key] = underlay_def
        return cast("UnderlayDefinition", underlay_def)

    def add_geodata(self, owner: str = "0", dxfattribs=None) -> GeoData:
        """Creates a new :class:`GeoData` entity and replaces existing ones.
        The GEODATA entity resides in the OBJECTS section and NOT in the layout
        entity space, and it is linked to the layout by an extension dictionary
        located in BLOCK_RECORD of the layout.

        The GEODATA entity requires DXF version R2010+. The DXF Reference does
        not document if other layouts than model space supports geo referencing,
        so getting/setting geo data may only make sense for the model space
        layout, but it is also available in paper space layouts.

        Args:
            owner: handle to owner as hex string
            dxfattribs: DXF attributes for :class:`~ezdxf.entities.GeoData` entity

        """
        if dxfattribs is None:
            dxfattribs = {}
        dxfattribs["owner"] = owner
        return cast("GeoData", self.add_dxf_object_with_reactor("GEODATA", dxfattribs))

    def reorg(self, auditor: Optional[Auditor] = None) -> Auditor:
        """Validate and recreate the integrity of the OBJECTS section."""
        if auditor is None:
            assert self.doc is not None, "valid document required"
            auditor = Auditor(self.doc)
        sanitizer = _Sanitizer(auditor, self)
        sanitizer.execute()
        return auditor


_OBJECT_TABLE_NAMES = [
    "ACAD_COLOR",
    "ACAD_GROUP",
    "ACAD_LAYOUT",
    "ACAD_MATERIAL",
    "ACAD_MLEADERSTYLE",
    "ACAD_MLINESTYLE",
    "ACAD_PLOTSETTINGS",
    "ACAD_PLOTSTYLENAME",
    "ACAD_SCALELIST",
    "ACAD_TABLESTYLE",
    "ACAD_VISUALSTYLE",
]

KNOWN_DICT_CONTENT: dict[str, str] = {
    "ACAD_COLOR": "DBCOLOR",
    "ACAD_GROUP": "GROUP",
    "ACAD_IMAGE_DICT": "IMAGEDEF",
    "ACAD_DETAILVIEWSTYLE": "ACDBDETAILVIEWSTYLE",
    "ACAD_LAYOUT": "LAYOUT",
    "ACAD_MATERIAL": "MATERIAL",
    "ACAD_MLEADERSTYLE": "MLEADERSTYLE",
    "ACAD_MLINESTYLE": "MLINESTYLE",
    "ACAD_PLOTSETTINGS": "PLOTSETTINGS",
    "ACAD_RENDER_ACTIVE_SETTINGS": "MENTALRAYRENDERSETTINGS",
    "ACAD_SCALELIST": "SCALE",
    "ACAD_SECTIONVIEWSTYLE": "ACDBSECTIONVIEWSTYLE",
    "ACAD_TABLESTYLE": "TABLESTYLE",
    "ACAD_PDFDEFINITIONS": "PDFDEFINITION",
    "ACAD_DWFDEFINITIONS": "DWFDEFINITION",
    "ACAD_DGNDEFINITIONS": "DGNDEFINITION",
    "ACAD_VISUALSTYLE": "VISUALSTYLE",
    "AcDbVariableDictionary": "DICTIONARYVAR",
}


class _Sanitizer:
    def __init__(self, auditor: Auditor, objects: ObjectsSection) -> None:
        self.objects = objects
        self.auditor = auditor
        self.rootdict = objects.rootdict
        self.removed_entity = True

    def dictionaries(self) -> Iterator[Dictionary]:
        for d in self.objects:
            if isinstance(d, Dictionary):
                yield d

    def execute(self, max_loops=100) -> None:
        self.restore_owner_handles_of_dictionary_entries()
        self.validate_known_dictionaries()
        loops = 0
        self.removed_entity = True
        while self.removed_entity and loops < max_loops:
            loops += 1
            self.removed_entity = False
            self.remove_orphaned_dictionaries()
            # Run audit on all entities of the OBJECTS section to take the removed
            # dictionaries into account.
            self.audit_objects()
        self.create_required_structures()

    def restore_owner_handles_of_dictionary_entries(self) -> None:
        def reclaim_entity() -> None:
            entity.dxf.owner = dict_handle
            self.auditor.fixed_error(
                AuditError.INVALID_OWNER_HANDLE,
                f"Fixed invalid owner handle of {entity}.",
            )

        def purge_key():
            purge.append(key)
            self.auditor.fixed_error(
                AuditError.INVALID_OWNER_HANDLE,
                f"Removed invalid key {key} in {str(dictionary)}.",
            )

        entitydb = self.auditor.entitydb
        for dictionary in self.dictionaries():
            purge: list[str] = []  # list of keys to discard
            dict_handle = dictionary.dxf.handle
            for key, entity in dictionary.items():
                if isinstance(entity, str):
                    # handle is not resolved -> entity does not exist
                    purge_key()
                    continue
                owner_handle = entity.dxf.get("owner")
                if owner_handle == dict_handle:
                    continue
                parent_dict = entitydb.get(owner_handle)
                if isinstance(parent_dict, Dictionary) and parent_dict.find_key(entity):
                    # entity belongs to parent_dict, discard key
                    purge_key()
                else:
                    reclaim_entity()
            for key in purge:
                dictionary.discard(key)

    def remove_orphaned_dictionaries(self) -> None:
        def kill_dictionary():
            entitydb.discard(dictionary)
            dictionary._silent_kill()

        entitydb = self.auditor.entitydb
        rootdict = self.rootdict
        for dictionary in self.dictionaries():
            if dictionary is rootdict:
                continue
            owner = entitydb.get(dictionary.dxf.get("owner"))
            if owner is None:
                # owner does not exist:
                # A DICTIONARY without an owner has no purpose and the owner can not be
                # determined, except for searching all dictionaries for an entry that
                # references this DICTIONARY, this is done in the method
                # restore_owner_handles_of_dictionary_entries().
                kill_dictionary()
                continue
            if not isinstance(owner, Dictionary):
                continue
            key = owner.find_key(dictionary)
            if not key:  # owner dictionary has no entry for this dict
                kill_dictionary()

    def validate_known_dictionaries(self) -> None:
        from ezdxf.entities import DXFEntity

        auditor = self.auditor
        for dict_name, expected_type in KNOWN_DICT_CONTENT.items():
            object_dict = self.rootdict.get(dict_name)
            if not isinstance(object_dict, Dictionary):
                continue
            purge_keys: list[str] = []
            for key, entry in object_dict.items():
                if isinstance(entry, DXFEntity) and entry.dxftype() != expected_type:
                    auditor.fixed_error(
                        AuditError.REMOVED_INVALID_DXF_OBJECT,
                        f"Removed invalid type {entry} from {object_dict}<{dict_name}>, "
                        f"expected type {expected_type}",
                    )
                    purge_keys.append(key)
                    auditor.trash(entry)
            for key in purge_keys:
                object_dict.discard(key)

    def create_required_structures(self):
        self.objects.setup_object_management_tables(self.rootdict)
        doc = self.objects.doc
        # update ObjectCollections:
        doc.materials.update_object_dict()
        doc.materials.create_required_entries()
        doc.mline_styles.update_object_dict()
        doc.mline_styles.create_required_entries()
        doc.mleader_styles.update_object_dict()
        doc.mleader_styles.create_required_entries()
        doc.groups.update_object_dict()

    def cleanup_entitydb(self):
        self.auditor.empty_trashcan()
        self.auditor.entitydb.purge()

    def audit_objects(self):
        self.cleanup_entitydb()
        auditor = self.auditor
        current_db_size = len(auditor.entitydb)

        for entity in self.objects:
            if not entity.dxf.hasattr("owner"):
                # check_owner_exist() ignores entities without owner handle
                auditor.fixed_error(
                    code=AuditError.INVALID_OWNER_HANDLE,
                    message=f"Deleted {str(entity)} entity without owner handle.",
                )
                auditor.trash(entity)
                continue
            auditor.check_owner_exist(entity)
            entity.audit(auditor)

        auditor.empty_trashcan()
        if current_db_size != len(auditor.entitydb):
            self.removed_entity = True



# Copyright (c) 2018-2019, Manfred Moitzi
# License: MIT License
import os
import pytest
import ezdxf
from ezdxf.lldxf.const import versions_supported_by_new


@pytest.fixture(params=versions_supported_by_new)
def drawing(request):
    return ezdxf.new(request.param)


def create_block(dwg):
    # first create a block
    flag = dwg.blocks.new(name="FLAG")

    # add dxf entities to the block (the flag)
    # use basepoint = (x, y) to define an other basepoint than (0, 0)
    flag_symbol = [(0, 0), (0, 5), (4, 3), (0, 3)]
    flag.add_polyline2d(flag_symbol)
    flag.add_circle((0, 0), 0.4, dxfattribs={"color": 2})

    # define some attributes
    flag.add_attdef("NAME", (0.5, -0.5), {"height": 0.5, "color": 3})
    flag.add_attdef("XPOS", (0.5, -1.0), {"height": 0.25, "color": 4})
    flag.add_attdef("YPOS", (0.5, -1.5), {"height": 0.25, "color": 4})


def insert_block(layout):
    point = (10, 12)
    values = {
        "NAME": "REFNAME",
        "XPOS": "x = %.3f" % point[0],
        "YPOS": "y = %.3f" % point[1],
    }
    scale = 1.75
    layout.add_auto_blockref(
        "FLAG",
        point,
        values,
        dxfattribs={
            "xscale": scale,
            "yscale": scale,
            "layer": "FLAGS",
            "rotation": -15,
        },
    )


def test_rotated_block(drawing, tmpdir):
    create_block(drawing)
    insert_block(drawing.modelspace())

    filename = str(
        tmpdir.join("rotated_block_reference_%s.dxf" % drawing.dxfversion)
    )
    try:
        drawing.saveas(filename)
    except ezdxf.DXFError as e:
        pytest.fail(
            "DXFError: {0} for DXF version {1}".format(
                str(e), drawing.dxfversion
            )
        )
    assert os.path.exists(filename)



# Copyright (c) 2019-2020 Manfred Moitzi
# License: MIT License
# created 2019-02-18
import pytest
import ezdxf
from ezdxf.lldxf.extendedtags import ExtendedTags
from ezdxf.entities import factory
from ezdxf.entities.factory import ENTITY_CLASSES


@pytest.fixture(scope="module")
def doc():
    return ezdxf.new()


def test_registered_structural_entities():
    assert "CLASS" in ENTITY_CLASSES
    assert "TABLE" in ENTITY_CLASSES
    assert "BLOCK" in ENTITY_CLASSES
    assert "ENDBLK" in ENTITY_CLASSES


def test_registered_table_entries():
    assert "LAYER" in ENTITY_CLASSES
    assert "LTYPE" in ENTITY_CLASSES
    assert "STYLE" in ENTITY_CLASSES
    assert "DIMSTYLE" in ENTITY_CLASSES
    assert "APPID" in ENTITY_CLASSES
    assert "UCS" in ENTITY_CLASSES
    assert "VIEW" in ENTITY_CLASSES
    assert "VPORT" in ENTITY_CLASSES
    assert "BLOCK_RECORD" in ENTITY_CLASSES


def test_new():
    e = factory.new("POINT")
    assert e.doc is None
    assert e.dxf.handle is None
    assert e.dxf.owner is None
    assert e.is_alive is True
    assert e.is_virtual is True


POINT = """0
POINT
5
FEFE
8
0
10
0.0
20
0.0
30
0.0
"""


def test_factory_load():
    tags = ExtendedTags.from_text(POINT)
    e = factory.load(tags)
    assert e.dxftype() == "POINT"
    assert e.doc is None
    assert e.dxf.handle == "FEFE"
    assert e.dxf.owner is None
    assert e.is_alive is True
    assert e.is_virtual is True


def test_bind_entity_to_doc(doc):
    e = factory.new("POINT")
    factory.bind(e, doc)
    assert e.doc is doc
    assert e.dxf.handle is not None, "should have a handle"
    assert (
        e.dxf.handle in doc.entitydb
    ), "should be stored in the entity database"
    assert e.dxf.owner is None, "should not be linked to a layout or owner"
    assert e.is_virtual is False, "is not a virtual entity"


def test_bind_entity_with_existing_handle_to_doc(doc):
    e = factory.new("POINT")
    e.dxf.handle = "ABBA"
    factory.bind(e, doc)
    assert e.doc is doc
    assert e.dxf.handle == "ABBA", "should have the original handle"
    assert (
        e.dxf.handle in doc.entitydb
    ), "should be stored in the entity database"


def test_bind_dead_entity_to_doc(doc):
    e = factory.new("POINT")
    e.destroy()
    with pytest.raises(AssertionError):
        factory.bind(e, doc)


def test_is_bound_true(doc):
    e = factory.new("POINT")
    factory.bind(e, doc)
    assert factory.is_bound(e, doc) is True
    assert e.is_bound is True


def test_is_bound_false(doc):
    e = factory.new("POINT")
    assert factory.is_bound(e, doc) is False
    assert e.is_bound is False


def test_if_destroyed_entity_is_bound(doc):
    e = factory.new("POINT")
    factory.bind(e, doc)
    e.destroy()
    assert factory.is_bound(e, doc) is False
    assert e.is_bound is False


def test_create_db_entry(doc):
    e = factory.create_db_entry("POINT", {}, doc)
    assert e.doc is doc
    assert e.dxf.handle is not None, "should have a handle"
    assert (
        e.dxf.handle in doc.entitydb
    ), "should be stored in the entity database"
    assert e.dxf.owner is None, "should not be linked to a layout or owner"
    assert e.is_virtual is False, "is not a virtual entity"


def test_unbind_bound_entity(doc):
    e = factory.create_db_entry("POINT", {}, doc)
    doc.modelspace().add_entity(e)
    factory.unbind(e)
    assert e.is_alive, "should not be destroyed"
    assert e.is_virtual, "should be virtual entity"
    assert e.doc is None
    assert e.dxf.owner is None
    assert e.dxf.handle is None


def test_unbind_unbound_entity(doc):
    e = factory.new("POINT")
    # should not raise an exception
    factory.unbind(e)
    assert e.is_alive, "should not be destroyed"


def test_unbind_destroyed_entity(doc):
    e = factory.new("POINT")
    e.destroy()
    # should not raise an exception
    factory.unbind(e)
    assert e.is_alive is False



# Copyright (c) 2019-2021 Manfred Moitzi
# License: MIT License
import pytest
from ezdxf.entities.leader import Leader
from ezdxf.lldxf.tagwriter import TagCollector, basic_tags_from_text
from ezdxf.layouts import VirtualLayout
from ezdxf.protocols import SupportsVirtualEntities, query_virtual_entities

LEADER = """0
LEADER
5
0
330
0
100
AcDbEntity
8
0
100
AcDbLeader
3
DIMSTYLE
73
3
40
1.0
41
1.0
76
3
10
0.0
20
0.0
30
0.0
10
0.0
20
0.0
30
0.0
10
0.0
20
0.0
30
0.0
340
FEFE
210
0.0
220
0.0
230
1.0
213
0.0
223
0.0
233
0.0
"""


@pytest.fixture
def entity():
    return Leader.from_text(LEADER)


def test_registered():
    from ezdxf.entities.factory import ENTITY_CLASSES

    assert "LEADER" in ENTITY_CLASSES


def test_default_init():
    entity = Leader()
    assert entity.dxftype() == "LEADER"
    assert entity.dxf.handle is None
    assert entity.dxf.owner is None


def test_default_new():
    entity = Leader.new(
        handle="ABBA",
        owner="0",
        dxfattribs={
            "color": 7,
        },
    )
    assert entity.dxf.layer == "0"
    assert entity.dxf.color == 7
    assert entity.dxf.dimstyle == "Standard"
    assert entity.dxf.has_arrowhead == 1
    assert entity.dxf.path_type == 0
    assert entity.dxf.annotation_type == 3
    assert entity.dxf.hookline_direction == 1
    assert entity.dxf.has_hookline == 1
    assert entity.dxf.text_height == 1
    assert entity.dxf.text_width == 1
    assert entity.dxf.block_color == 7
    assert entity.dxf.annotation_handle == "0"
    assert entity.dxf.normal_vector == (0, 0, 1)
    assert entity.dxf.horizontal_direction == (1, 0, 0)
    assert entity.dxf.leader_offset_block_ref == (0, 0, 0)
    assert entity.dxf.leader_offset_annotation_placement == (0, 0, 0)
    assert len(entity.vertices) == 0


def test_load_from_text(entity):
    assert entity.dxf.layer == "0"
    assert entity.dxf.color == 256, "default color is 256 (by layer)"
    assert entity.dxf.dimstyle == "DIMSTYLE"
    assert entity.dxf.has_arrowhead == 1
    assert entity.dxf.path_type == 0
    assert entity.dxf.annotation_type == 3
    assert entity.dxf.hookline_direction == 1
    assert entity.dxf.has_hookline == 1
    assert entity.dxf.text_height == 1
    assert entity.dxf.text_width == 1
    assert entity.dxf.block_color == 7
    assert entity.dxf.annotation_handle == "FEFE"
    assert entity.dxf.normal_vector == (0, 0, 1)
    assert entity.dxf.horizontal_direction == (1, 0, 0)
    assert entity.dxf.leader_offset_block_ref == (0, 0, 0)
    assert entity.dxf.leader_offset_annotation_placement == (0, 0, 0)
    assert len(entity.vertices) == 3


def test_write_dxf():
    entity = Leader.from_text(LEADER)
    result = TagCollector.dxftags(entity)
    expected = basic_tags_from_text(LEADER)
    assert result == expected


VERTICES = [
    (0, 0, 0),
    (1, 1, 0),
    (2, 1, 0),
]


def test_add_leader():
    msp = VirtualLayout()
    leader = msp.add_leader(vertices=VERTICES)
    assert leader.dxftype() == "LEADER"
    assert leader.dxf.annotation_type == 3
    assert len(leader.vertices) == 3
    assert leader.vertices == VERTICES


def test_supports_virtual_entities_protocol():
    leader = Leader.new()
    leader.vertices = VERTICES
    assert isinstance(leader, SupportsVirtualEntities)
    assert len(query_virtual_entities(leader)) == 2


def test_virtual_sub_entities_source_tracking():
    leader = Leader.new()
    leader.vertices = VERTICES

    result = set(e.source_of_copy for e in leader.virtual_entities())
    assert len(result) == 1, "only one source of copy expected"
    assert leader in result, "lwpolyline should be the source of copy"



# Copyright (c) 2018 Manfred Moitzi
# License: MIT License
import math
import os

import pytest

import ezdxf
from ezdxf.entities.geodata import GeoData, EPSG_3395
from ezdxf.lldxf.tagwriter import TagCollector, basic_tags_from_text
from ezdxf.math import Vec3, Matrix44

GEODATA = """0
GEODATA
5
0
102
{ACAD_REACTORS
330
0
102
}
330
DEAD
100
AcDbGeoData
90
2
330
70
70
2
10
0.0
20
0.0
30
0.0
11
0.0
21
0.0
31
0.0
40
1.0
91
1
41
1.0
92
1
210
0.0
220
0.0
230
1.0
12
0
22
1
95
3
141
1.0
294
0
142
0.0
143
0.0
301
COORDINATE_SYSTEM_DEFINITION
93
0
96
0
"""


@pytest.fixture
def entity():
    return GeoData.from_text(GEODATA)


def test_registered():
    from ezdxf.entities.factory import ENTITY_CLASSES

    assert "GEODATA" in ENTITY_CLASSES


def test_default_init():
    entity = GeoData()
    assert entity.dxftype() == "GEODATA"
    assert entity.dxf.handle is None
    assert entity.dxf.owner is None


def test_default_new():
    entity = GeoData.new(handle="ABBA", owner="0", dxfattribs={})
    assert entity.dxf.version == 2
    assert entity.dxf.block_record_handle == "0"
    assert entity.dxf.design_point == (0, 0, 0)
    assert entity.dxf.reference_point == (0, 0, 0)
    assert entity.dxf.north_direction == (0, 1)
    assert entity.dxf.horizontal_unit_scale == 1
    assert entity.dxf.vertical_unit_scale == 1
    assert entity.dxf.horizontal_units == 1
    assert entity.dxf.vertical_units == 1
    assert entity.dxf.up_direction == (0, 0, 1)
    assert entity.dxf.scale_estimation_method == 1
    assert entity.dxf.sea_level_correction == 0
    assert entity.dxf.user_scale_factor == 1
    assert entity.dxf.sea_level_elevation == 0
    assert entity.dxf.coordinate_projection_radius == 0
    assert entity.dxf.geo_rss_tag == ""
    assert entity.dxf.observation_from_tag == ""
    assert entity.dxf.observation_to_tag == ""
    assert len(entity.source_vertices) == 0
    assert len(entity.target_vertices) == 0
    assert len(entity.faces) == 0
    assert entity.coordinate_system_definition == ""


def test_load_from_text(entity):
    assert entity.dxf.version == 2
    assert entity.dxf.block_record_handle == "70"
    assert entity.dxf.design_point == (0, 0, 0)
    assert entity.dxf.reference_point == (0, 0, 0)
    assert entity.dxf.north_direction == (0, 1)
    assert entity.dxf.horizontal_unit_scale == 1
    assert entity.dxf.vertical_unit_scale == 1
    assert entity.dxf.horizontal_units == 1
    assert entity.dxf.vertical_units == 1
    assert entity.dxf.up_direction == (0, 0, 1)
    assert entity.dxf.scale_estimation_method == 3
    assert entity.dxf.sea_level_correction == 0
    assert entity.dxf.user_scale_factor == 1
    assert entity.dxf.sea_level_elevation == 0
    assert entity.dxf.coordinate_projection_radius == 0
    assert entity.dxf.geo_rss_tag == ""
    assert entity.dxf.observation_from_tag == ""
    assert entity.dxf.observation_to_tag == ""
    assert len(entity.source_vertices) == 0
    assert len(entity.target_vertices) == 0
    assert len(entity.faces) == 0
    assert entity.coordinate_system_definition == "COORDINATE_SYSTEM_DEFINITION"


def test_write_dxf():
    entity = GeoData.from_text(GEODATA)
    result = TagCollector.dxftags(entity)
    expected = basic_tags_from_text(GEODATA)
    assert result == expected


@pytest.fixture
def geodata():
    return GeoData.from_text(GEODATA2)


def test_geodata_dxf_attributes(geodata):
    assert geodata.dxf.handle == "F658"
    assert geodata.has_reactors() is True
    assert geodata.dxf.version == 3
    assert geodata.dxf.block_record_handle == "70"
    assert geodata.dxf.design_point == (-7000, -1000, 0)
    assert geodata.dxf.reference_point == (14, 48, 0)
    assert geodata.dxf.north_direction == (0.13, 0.99)
    assert geodata.dxf.horizontal_unit_scale == 1.0
    assert geodata.dxf.vertical_unit_scale == 1.0
    assert geodata.dxf.horizontal_units == 6
    assert geodata.dxf.vertical_units == 6
    assert geodata.dxf.up_direction == (0, 0, 1)
    assert geodata.dxf.scale_estimation_method == 3
    assert geodata.dxf.sea_level_correction == 0
    assert geodata.dxf.user_scale_factor == 1
    assert geodata.dxf.sea_level_elevation == 0
    assert geodata.dxf.coordinate_projection_radius == 0
    assert geodata.dxf.geo_rss_tag == "DUMMY_TAG"
    assert geodata.dxf.observation_from_tag == "DUMMY_OFT"
    assert geodata.dxf.observation_to_tag == "DUMMY_OTT"
    assert len(geodata.source_vertices) == 4


def test_geodata_get_mesh_data(geodata):
    assert len(geodata.source_vertices) == 4
    assert len(geodata.target_vertices) == 4
    assert len(geodata.faces) == 0


def test_geodata_delete_and_extend_mesh_data(geodata):
    geodata.source_vertices.clear()
    geodata.target_vertices.clear()
    assert len(geodata.source_vertices) == 0
    assert len(geodata.target_vertices) == 0
    geodata.source_vertices.extend([(1, 1), (2, 2), (3, 3)])
    geodata.target_vertices.extend([(1, 1), (2, 2), (3, 3)])
    assert len(geodata.source_vertices) == 3
    assert len(geodata.target_vertices) == 3


def test_geodata_set_mesh_data(geodata):
    # vertex structure
    source_vertices = [(1, 1), (2, 2), (3, 3)]
    target_vertices = [(4, 4), (5, 5), (6, 6)]
    geodata.source_vertices.set(source_vertices)
    geodata.target_vertices.set(target_vertices)
    # faces is a standard python list
    geodata.faces.append([0, 1, 2])
    assert len(geodata.source_vertices) == 3
    assert len(geodata.target_vertices) == 3
    assert len(geodata.faces) == 1


TEST_TEXT = """Lorem ipsum dolor sit amet, consetetur sadipscing elitr, sed diam nonumy eirmod tempor invidunt ut labore et dolore 
magna aliquyam erat, sed diam voluptua. At vero eos et accusam et justo duo dolores et ea rebum. Stet clita kasd 
gubergren, no sea takimata sanctus est Lorem ipsum dolor sit amet. Lorem ipsum dolor sit amet, consetetur sadipscing 
elitr, sed diam nonumy eirmod tempor invidunt ut labore et dolore magna aliquyam erat, sed diam voluptua. At vero 
eos et accusam et justo duo dolores et ea rebum. Stet clita kasd gubergren, no sea takimata sanctus est Lorem ipsum 
dolor sit amet.
"""


def test_geodata_coordinate_system_definition(geodata):
    assert geodata.coordinate_system_definition == "Text0Text1"

    geodata.coordinate_system_definition = TEST_TEXT
    collector = TagCollector()
    geodata.export_coordinate_system_definition(collector)
    tags = collector.tags
    assert len(tags) == 3
    assert tags[0].code == 303
    # \n converted to ^J
    assert "^J" in tags[0].value
    assert tags[1].code == 303
    assert tags[2].code == 301


def test_create_new_geo_data_for_model_space():
    doc = ezdxf.new("R2010")
    msp = doc.modelspace()
    assert msp.get_geodata() is None
    geodata = msp.new_geodata()
    assert geodata.dxftype() == "GEODATA"


GEODATA2 = """  0
GEODATA
5
F658
102
{ACAD_REACTORS
330
803A
102
}
330
803A
100
AcDbGeoData
90
3
330
70
70
2
10
-7000
20
-1000
30
0.0
11
14.
21
48.
31
0.0
40
1.0
91
6
41
1.0
92
6
210
0.0
220
0.0
230
1.0
12
0.13
22
0.99
95
3
141
1.0
294
0
142
0.0
143
0.0
303
Text0
301
Text1
302
DUMMY_TAG
305
DUMMY_OFT
306
DUMMY_OTT
307

93
4
13
-596044.3104449062
23
-1966225.111236282
14
17.66571737167506
24
42.00025909173485
13
1403002.41871386
23
-1966225.111236282
14
41.46357474182291
24
40.89934076070281
13
1403002.41871386
23
-339583.1428681399
14
46.99412402090528
24
55.00032933828663
13
-596044.3104449062
23
-337956.5008997718
14
15.23395535970868
24
56.52797613716202
96
0
"""


@pytest.fixture
def georeferenced_test_file_path() -> str:
    return os.path.join(
        os.path.dirname(__file__), "houses_of_parliament_georeferenced.dxf"
    )


def test_interpreting_geodata(georeferenced_test_file_path):
    # It is unclear how to create a georeferenced file from scratch. Copying
    #     # every GeoData attribute and document header value across was not enough
    #     # for AutoCAD to correctly interpret the coordinates.
    doc = ezdxf.readfile(georeferenced_test_file_path)
    geodata = doc.modelspace().get_geodata()

    assert geodata.decoded_units() == ("in", "in")  # inches

    coordinate_system_definition = geodata.coordinate_system_definition
    geodata.coordinate_system_definition = ""
    with pytest.raises(ezdxf.InvalidGeoDataException):
        geodata.get_crs()
    geodata.coordinate_system_definition = coordinate_system_definition

    assert geodata.get_crs() == (27700, True)

    # the outline of the Houses of Parliament in London, WCS coordinates are
    # meaningless, but the cad file is georeferenced as epsg:27700
    expected_geo_points = [
        (530207.5677217417, 179366.7895852687),
        (530304.7243275082, 179354.44795162012),
        (530337.0722795193, 179620.5081263125),
        (530285.502052045, 179626.77810356658),
        (530288.2604343889, 179649.46565077276),
        (530260.6778594478, 179652.81917713786),
        (530256.9625486559, 179629.5648430319),
        (530266.9241695277, 179627.97328950214),
        (530259.8925073204, 179577.1375901363),
        (530179.7793595218, 179598.3848030573),
        (530174.8000936891, 179519.37903558338),
        (530189.905505017, 179515.41460442453),
        (530184.6203071928, 179495.27676273056),
        (530217.9628916974, 179486.52596620753),
        (530223.1861964873, 179485.42147930583),
        (530207.5677217417, 179366.7895852687),
    ]
    expected_transformation = Matrix44(
        (0.02154042164237322, -0.013459949311523403, 0.0, 0.0),
        (0.013459949311523403, 0.02154042164237322, 0.0, 0.0),
        (0.0, 0.0, 1.0, 0.0),
        (529635.6280343985, 179575.4070305015, 0.0, 1.0),
    )

    transformation, epsg = geodata.get_crs_transformation()
    assert epsg == 27700
    assert all(
        math.isclose(x1, x2)
        for x1, x2 in zip(transformation, expected_transformation)
    )

    entity = list(doc.modelspace().query("LWPOLYLINE"))[0]
    georeferenced_entity = entity.transform(transformation)
    transformed_points = georeferenced_entity.get_points(format="xy")
    assert len(transformed_points) == len(expected_geo_points)
    assert all(
        Vec3(x1).isclose(Vec3(x2))
        for x1, x2 in zip(transformed_points, expected_geo_points)
    )


def test_setup_local_grid_epsg_3395():
    doc = ezdxf.new("R2010", units=5)  # cm
    msp = doc.modelspace()
    geodata = msp.new_geodata()
    geodata.setup_local_grid(
        design_point=(0, 0, 0), reference_point=(1718030, 5921664, 0)  # Graz
    )
    assert geodata.coordinate_system_definition is EPSG_3395
    assert geodata.dxf.design_point == (0, 0, 0)
    assert geodata.dxf.reference_point == (1718030, 5921664, 0)
    assert geodata.dxf.coordinate_type == geodata.LOCAL_GRID
    assert geodata.dxf.horizontal_units == 5
    assert geodata.dxf.horizontal_unit_scale == 0.01
    assert geodata.dxf.vertical_units == 5
    assert geodata.dxf.vertical_unit_scale == 0.01
    assert geodata.dxf.north_direction == (0, 1)
    assert geodata.dxf.up_direction == (0, 0, 1)
    assert geodata.dxf.scale_estimation_method == geodata.NONE



# Copyright (c) 2011-2022, Manfred Moitzi
# License: MIT License
import pytest
import ezdxf
from ezdxf.sections.tables import TablesSection
from ezdxf.lldxf.tagwriter import TagCollector


class TestGenericTableFeatures:
    @pytest.fixture(scope="class")
    def tables(self) -> TablesSection:
        doc = ezdxf.new()
        return doc.tables

    def test_constructor(self, tables):
        assert tables.layers is not None
        assert tables.linetypes is not None
        assert tables.appids is not None
        assert tables.styles is not None
        assert tables.dimstyles is not None
        assert tables.views is not None
        assert tables.viewports is not None
        assert tables.ucs is not None
        assert tables.block_records is not None

    def test_getattr_upper_case(self, tables):
        with pytest.raises(AttributeError):
            _ = tables.LINETYPES

    def test_error_getattr(self, tables):
        with pytest.raises(AttributeError):
            _ = tables.test

    def test_get_entry_by_handle(self, tables):
        layer = tables.layers.add("Test0")
        assert tables.layers.get_entry_by_handle(layer.dxf.handle) is layer

    def test_get_non_existing_entry_by_handle(self, tables):
        assert tables.layers.get_entry_by_handle("FEFE") is None

    def test_get_handle_of_entry(self, tables):
        layer = tables.layers.add("Test1")
        assert tables.layers.get_handle_of_entry("test1") == layer.dxf.handle

    def test_get_handle_of_non_existing_entry(self, tables):
        assert tables.layers.get_handle_of_entry("test2") == ""

    def test_get_entry_by_handle_is_type_safe(self, tables):
        arial = tables.styles.add("Arial0", font="arial.ttf")
        assert tables.layers.get_entry_by_handle(arial.dxf.handle) is None


class TestLayerTableEntry:
    @pytest.fixture(scope="class")
    def tables(self) -> TablesSection:
        doc = ezdxf.new()
        return doc.tables

    def test_add_layer(self, tables):
        layer = tables.layers.add(
            "NEW_LAYER",
            color=2,
            true_color=ezdxf.rgb2int((0x10, 0x20, 0x30)),
            linetype="DASHED",
            lineweight=18,
            plot=True,
        )
        assert layer.dxf.name == "NEW_LAYER"
        assert layer.dxf.color == 2
        assert layer.dxf.true_color == 0x00102030
        assert layer.dxf.linetype == "DASHED", "no check if line type exist!"
        assert layer.dxf.lineweight == 18
        assert layer.dxf.plot == 1

    def test_check_invalid_aci_color(self, tables):
        with pytest.raises(ValueError):
            tables.layers.add("INVALID_ACI", color=300)

    def test_check_invalid_line_weight(self, tables):
        with pytest.raises(ValueError):
            tables.layers.add("INVALID_LINE_WEIGHT", lineweight=300)

    def test_add_new_line_type(self, tables):
        ltype = tables.linetypes.add(
            "SIMPLE_LINE_TYPE", [0.2, 0.1, -0.1], description="description"
        )
        assert ltype.dxf.name == "SIMPLE_LINE_TYPE"
        assert ltype.dxf.description == "description"
        # Correct pattern creation is tested in test suite 121.


class TestTextStyleTable:
    @pytest.fixture(scope="class")
    def tables(self) -> TablesSection:
        doc = ezdxf.new()
        return doc.tables

    def test_add_new_ttf_font_text_style(self, tables):
        style = tables.styles.add(
            "NEW_STYLE", font="Arial.ttf", dxfattribs={"flags": 3}
        )
        assert style.dxf.name == "NEW_STYLE"
        assert style.dxf.font == "Arial.ttf"
        assert style.dxf.flags == 3

    def test_add_new_shape_file(self, tables):
        style = tables.styles.add_shx("shapes1.shx")
        assert style.dxf.name == "", "shape files have no name"
        assert style.dxf.font == "shapes1.shx"
        assert style.dxf.flags == 1
        assert style.is_shape_file is True

        # can not add same shape file twice:
        with pytest.raises(ezdxf.const.DXFTableEntryError):
            tables.styles.add_shx("shapes1.shx")

    def test_add_multiple_shape_files(self, tables):
        style1 = tables.styles.add_shx("shapes4.shx")
        style2 = tables.styles.add_shx("shapes5.shx")
        assert tables.styles.find_shx(style1.dxf.font) is not None
        assert tables.styles.find_shx(style2.dxf.font) is not None

    def test_get_shape_file(self, tables):
        style = tables.styles.get_shx("shapes2.shx")
        assert style.dxf.name == "", "shape files have no name"
        assert style.dxf.font == "shapes2.shx"
        assert style.dxf.flags == 1

        style2 = tables.styles.get_shx("shapes2.shx")
        assert style is style2

    def test_find_shape_file(self, tables):
        tables.styles.add_shx("shapes3.shx")
        style = tables.styles.find_shx("shapes3.shx")
        assert style.dxf.font == "shapes3.shx"

    def test_if_shape_file_entry_exist(self, tables):
        assert tables.styles.find_shx("unknown.shx") is None

    def test_discard_shape_files(self, tables: TablesSection):
        style1 = tables.styles.add_shx("shapes6.shx")
        style2 = tables.styles.add_shx("shapes7.shx")
        tables.styles.discard_shx("shapes6.shx")
        tables.styles.discard_shx("shapes7.shx")
        assert tables.styles.find_shx(style1.dxf.font) is None
        assert tables.styles.find_shx(style2.dxf.font) is None

    def test_discard_not_existing_shape_file(self, tables):
        assert tables.styles.find_shx("shapes8.shx") is None
        tables.styles.discard_shx("shapes8.shx")  # should not raise an exception

    def test_export_multiple_shape_file_entries(self, tables):
        doc = ezdxf.new()
        doc.styles.add_shx("export1.shx")
        doc.styles.add_shx("export2.shx")

        tagwriter = TagCollector()
        doc.styles.export_dxf(tagwriter)
        font_names = [tag.value for tag in tagwriter.tags if tag.code == 3]
        assert len(font_names) == 3
        assert "export1.shx" in font_names
        assert "export2.shx" in font_names
        assert "txt" in font_names, "expected font of default text style STANDARD"


class TestLineTypeTable:
    @pytest.fixture(scope="class")
    def tables(self) -> TablesSection:
        doc = ezdxf.new()
        return doc.tables

    def test_add_new_line_type(self, tables):
        ltype = tables.linetypes.add(
            "SIMPLE_LINE_TYPE", [0.2, 0.1, -0.1], description="description"
        )
        assert ltype.dxf.name == "SIMPLE_LINE_TYPE"
        assert ltype.dxf.description == "description"
        # Correct pattern creation is tested in test suite 121.



# Copyright (c) 2018-2022, Manfred Moitzi
# License: MIT License
import pytest
import math
import ezdxf
from ezdxf.math import Vec3, BoundingBox, Matrix44, close_vectors
from ezdxf.render import forms
from ezdxf.addons.menger_sponge import MengerSponge
from ezdxf.render.mesh import (
    MeshVertexMerger,
    MeshBuilder,
    MeshTransformer,
    MeshAverageVertexMerger,
    merge_connected_paths,
    merge_full_patch,
    NodeMergingError,
    DegeneratedPathError,
    remove_colinear_face_vertices,
    all_edges,
    get_edge_stats,
    separate_meshes,
    face_normals_after_transformation,
    FaceOrientationDetector,
    volume6,
)
from ezdxf.addons import SierpinskyPyramid
from ezdxf.layouts import VirtualLayout


def test_vertex_merger_indices():
    merger = MeshVertexMerger()
    indices = merger.add_vertices([(1, 2, 3), (4, 5, 6)])
    indices2 = merger.add_vertices([(1, 2, 3), (4, 5, 6)])
    assert indices == indices2


def test_vertex_merger_vertices():
    merger = MeshVertexMerger()
    merger.add_vertices([(1, 2, 3), (4, 5, 6)])
    merger.add_vertices([(1, 2, 3), (4, 5, 6)])
    assert merger.vertices == [(1, 2, 3), (4, 5, 6)]


def test_vertex_merger_index_of():
    merger = MeshVertexMerger()
    merger.add_vertices([(1, 2, 3), (4, 5, 6)])
    assert merger.index((1, 2, 3)) == 0
    assert merger.index((4, 5, 6)) == 1
    with pytest.raises(IndexError):
        merger.index((7, 8, 9))


def test_average_vertex_merger_indices():
    merger = MeshAverageVertexMerger()
    indices = merger.add_vertices([(1, 2, 3), (4, 5, 6)])
    indices2 = merger.add_vertices([(1, 2, 3), (4, 5, 6)])
    assert indices == indices2


def test_average_vertex_merger_vertices():
    merger = MeshAverageVertexMerger()
    merger.add_vertices([(1, 2, 3), (4, 5, 6)])
    merger.add_vertices([(1, 2, 3), (4, 5, 6)])
    assert merger.vertices == [(1, 2, 3), (4, 5, 6)]


def test_average_vertex_merger_index_of():
    merger = MeshAverageVertexMerger()
    merger.add_vertices([(1, 2, 3), (4, 5, 6)])
    assert merger.index((1, 2, 3)) == 0
    assert merger.index((4, 5, 6)) == 1
    with pytest.raises(IndexError):
        merger.index((7, 8, 9))


def test_mesh_builder(msp):
    pyramid = SierpinskyPyramid(level=4, sides=3)
    pyramid.render(msp, merge=False)
    meshes = msp.query("MESH")
    assert len(meshes) == 256


def test_vertex_merger():
    pyramid = SierpinskyPyramid(level=4, sides=3)
    faces = pyramid.faces()
    mesh = MeshVertexMerger()
    for vertices in pyramid:
        mesh.add_mesh(vertices=vertices, faces=faces)
    assert len(mesh.vertices) == 514
    assert len(mesh.faces) == 1024


def test_average_vertex_merger():
    pyramid = SierpinskyPyramid(level=4, sides=3)
    faces = pyramid.faces()
    mesh = MeshAverageVertexMerger()
    for vertices in pyramid:
        mesh.add_mesh(vertices=vertices, faces=faces)
    assert len(mesh.vertices) == 514
    assert len(mesh.faces) == 1024


REGULAR_FACE = Vec3.list([(0, 0, 0), (1, 0, 1), (1, 1, 1), (0, 1, 0)])
IRREGULAR_FACE = Vec3.list([(0, 0, 0), (1, 0, 1), (1, 1, 0), (0, 1, 0)])


def test_has_none_planar_faces():
    mesh = MeshBuilder()
    mesh.add_face(REGULAR_FACE)
    assert mesh.diagnose().has_non_planar_faces() is False
    mesh.add_face(IRREGULAR_FACE)
    assert mesh.diagnose().has_non_planar_faces() is True


def test_scale_mesh():
    mesh = forms.cube(center=False)
    mesh.scale(2, 3, 4)
    bbox = BoundingBox(mesh.vertices)
    assert bbox.extmin.isclose((0, 0, 0))
    assert bbox.extmax.isclose((2, 3, 4))


def test_rotate_x():
    mesh = forms.cube(center=False)
    mesh.rotate_x(math.radians(90))
    bbox = BoundingBox(mesh.vertices)
    assert bbox.extmin.isclose((0, -1, 0))
    assert bbox.extmax.isclose((1, 0, 1))


def test_mesh_bounding_box():
    bbox = forms.cube().bbox()
    assert bbox.extmin.isclose((-0.5, -0.5, -0.5))
    assert bbox.extmax.isclose((0.5, 0.5, 0.5))


def test_get_face_vertices():
    mesh = forms.cube(center=False)
    vertices = mesh.get_face_vertices(0)  # downward bottom face!
    assert close_vectors(vertices, [(0, 0), (0, 1), (1, 1), (1, 0)]) is True


def test_index_error_for_getting_face_vertices():
    mesh = forms.cube(center=False)
    with pytest.raises(IndexError):
        mesh.get_face_vertices(10)


def test_get_face_normal():
    mesh = forms.cube(center=False)
    assert mesh.get_face_normal(0).isclose((0, 0, -1)), "downward bottom face"
    assert mesh.get_face_normal(-1).isclose((0, 0, 1)), "upward top face"


class TestMeshDiagnose:
    def test_empty_mesh_is_not_watertight(self):
        mesh = MeshBuilder()
        assert mesh.diagnose().euler_characteristic != 2

    def test_single_face_mesh_is_not_watertight(self):
        mesh = MeshBuilder()
        mesh.add_face(REGULAR_FACE)
        assert mesh.diagnose().euler_characteristic != 2

    def test_cube_is_watertight(self):
        mesh = forms.cube(center=False)
        assert mesh.diagnose().euler_characteristic == 2

    def test_is_watertight_can_not_detect_vertex_orientation_errors(self):
        mesh = forms.cube(center=False)
        mesh.faces[-1] = tuple(reversed(mesh.faces[-1]))
        assert mesh.diagnose().euler_characteristic == 2

    def test_edge_balance_of_closed_surface_is_not_broken(self):
        mesh = forms.cube(center=False)
        assert mesh.diagnose().is_edge_balance_broken is False

    def test_edge_balance_of_wrong_oriented_faces_is_broken(self):
        mesh = forms.cube(center=False)
        mesh.faces[-1] = tuple(reversed(mesh.faces[-1]))
        assert mesh.diagnose().is_edge_balance_broken is True

    def test_edge_balance_of_doubled_faces_is_broken(self):
        mesh = forms.cube(center=False)
        mesh.faces.append(mesh.faces[-1])
        assert mesh.diagnose().is_edge_balance_broken is True

    def test_total_edge_count_of_closed_surface(self):
        mesh = forms.cube(center=False)
        stats = mesh.diagnose()
        assert stats.total_edge_count() == stats.n_edges * 2

    def test_cube_of_separated_faces_is_not_watertight(self):
        mesh = forms.cube(center=False)
        mesh2 = MeshBuilder()
        for face in mesh.faces_as_vertices():
            mesh2.add_face(face)
        assert mesh2.diagnose().euler_characteristic != 2

    def test_cylinder_is_watertight(self):
        mesh = forms.cylinder()
        assert mesh.diagnose().euler_characteristic == 2

    @pytest.mark.parametrize(
        "surface",
        [forms.cube(), forms.cylinder(), forms.cone(), forms.sphere()],
        ids="cube cylinder cone sphere".split(),
    )
    def test_surface_normals_pointing_outwards(self, surface):
        diagnose = surface.diagnose()
        assert diagnose.estimate_face_normals_direction() > 0.9

    def test_cylinder_with_reversed_cap_normals(self):
        c = forms.cylinder()
        for i, face in enumerate(c.faces):
            if len(face) > 4:
                c.faces[i] = tuple(reversed(c.faces[i]))
        diagnose = c.diagnose()
        assert diagnose.estimate_face_normals_direction() < 0.8
        assert diagnose.is_edge_balance_broken is True

    def test_cube_is_manifold(self):
        diag = forms.cube().diagnose()
        assert diag.is_manifold is True

    def test_mesh_bounding_box(self):
        bbox = forms.cube().diagnose().bbox
        assert bbox.extmin.isclose((-0.5, -0.5, -0.5))
        assert bbox.extmax.isclose((0.5, 0.5, 0.5))

    def test_regular_open_surface(self):
        cube = forms.cube()
        cube.faces.pop()
        diag = cube.diagnose()
        assert diag.is_closed_surface is False

    def test_regular_cube_has_closed_surface(self):
        diag = forms.cube().diagnose()
        assert diag.is_closed_surface is True

    def test_multiple_cubes_have_closed_surface(self):
        cube = forms.cube()
        cube2 = cube.copy()
        cube2.translate(10)
        cube.add_mesh(mesh=cube2)
        diag = cube.diagnose()
        assert diag.is_closed_surface is True

    def test_cube_with_reversed_face_has_closed_surface(self):
        cube = forms.cube()
        cube.faces[0] = tuple(reversed(cube.faces[0]))
        diag = cube.diagnose()
        assert diag.is_closed_surface is True

    def test_non_manifold_cube_has_not_a_closed_surface(self):
        cube = forms.cube()
        cube.faces.append(cube.faces[0])
        diag = cube.diagnose()
        assert diag.is_closed_surface is False

    def test_get_face_normals(self):
        cube = forms.cube()
        assert len(cube.diagnose().face_normals) == 6


def test_flipped_cube_normals_pointing_inwards():
    c = forms.cube()
    c.flip_normals()
    diagnose = c.diagnose()
    assert diagnose.estimate_face_normals_direction() == pytest.approx(-1.0)


@pytest.fixture
def msp():
    return VirtualLayout()


@pytest.fixture(scope="module")
def cube_polyface():
    layout = VirtualLayout()
    p = layout.add_polyface()
    p.append_faces(forms.cube().faces_as_vertices())
    return p


def test_from_empty_polyface(msp):
    empty_polyface = msp.add_polyface()
    b = MeshBuilder.from_polyface(empty_polyface)
    assert len(b.vertices) == 0
    assert len(b.faces) == 0


def test_from_cube_polyface(cube_polyface):
    b = MeshBuilder.from_polyface(cube_polyface)
    assert len(b.vertices) == 24  # unoptimized mesh builder
    assert len(b.faces) == 6


def test_render_polyface(cube_polyface, msp):
    t = MeshTransformer.from_polyface(cube_polyface)
    assert len(t.vertices) == 24  # unoptimized mesh builder
    assert len(t.faces) == 6
    t.render_polyface(msp)
    new_polyface = msp[-1]
    assert new_polyface.dxftype() == "POLYLINE"
    assert new_polyface.is_poly_face_mesh is True
    assert len(new_polyface.vertices) == 8 + 6
    assert new_polyface.vertices[0] is not cube_polyface.vertices[0]


def test_render_3dsolid():
    """Test if the render_3dsolid() method works.

    The ACIS content is tested by the tests for the acis module 529-532.
    """
    # A valid DXF document is required!
    doc = ezdxf.new(ezdxf.DXF2000)
    msp = doc.modelspace()
    cube = forms.cube()
    solid3d = cube.render_3dsolid(msp)
    assert solid3d.dxftype() == "3DSOLID"


def test_from_polymesh(msp):
    polymesh = msp.add_polymesh(size=(4, 4))
    b = MeshBuilder.from_polyface(polymesh)
    n = polymesh.dxf.n_count
    m = polymesh.dxf.m_count
    nfaces = (n - 1) * (m - 1)
    assert len(b.vertices) == nfaces * 4  # unoptimized mesh builder
    assert len(b.faces) == nfaces


def test_from_polyface_type_error(msp):
    polyline = msp.add_polyline3d([(0, 0, 0), (1, 0, 0)])
    with pytest.raises(TypeError):
        MeshBuilder.from_polyface(polyline)

    line = msp.add_line(start=(0, 0, 0), end=(1, 0, 0))
    with pytest.raises(TypeError):
        MeshBuilder.from_polyface(line)


@pytest.fixture
def polyface_181_1(msp):
    e = msp.new_entity(
        "POLYLINE",
        dxfattribs={
            "flags": 48,
            "m_count": 2,
            "n_count": 6,
        },
    )
    e.append_vertex(
        (25041.94191089287, 29272.95055566061, 0.0), dxfattribs={"flags": 64}
    )
    e.append_vertex(
        (25020.29127589287, 29285.45055566061, 0.0), dxfattribs={"flags": 64}
    )
    e.append_vertex(
        (25020.29127589287, 29310.45055566061, 0.0), dxfattribs={"flags": 64}
    )
    e.append_vertex(
        (25041.94191089287, 29322.95055566061, 0.0), dxfattribs={"flags": 64}
    )
    e.append_vertex(
        (25063.59254589287, 29310.45055566061, 0.0), dxfattribs={"flags": 64}
    )
    e.append_vertex(
        (25063.59254589287, 29285.45055566061, 0.0), dxfattribs={"flags": 64}
    )
    e.append_vertex(
        (25041.94191089287, 29272.95055566061, 50.0), dxfattribs={"flags": 64}
    )
    e.append_vertex(
        (25020.29127589287, 29285.45055566061, 50.0), dxfattribs={"flags": 64}
    )
    e.append_vertex(
        (25020.29127589287, 29310.45055566061, 50.0), dxfattribs={"flags": 64}
    )
    e.append_vertex(
        (25041.94191089287, 29322.95055566061, 50.0), dxfattribs={"flags": 64}
    )
    e.append_vertex(
        (25063.59254589287, 29310.45055566061, 50.0), dxfattribs={"flags": 64}
    )
    e.append_vertex(
        (25063.59254589287, 29285.45055566061, 50.0), dxfattribs={"flags": 64}
    )
    return e


def test_from_polyface_182_1(polyface_181_1):
    mesh = MeshVertexMerger.from_polyface(polyface_181_1)
    assert len(mesh.vertices) == 12


@pytest.fixture
def polyface_181_2(msp):
    e = msp.new_entity(
        "POLYLINE",
        dxfattribs={
            "flags": 16,
            "m_count": 6,
            "n_count": 3,
        },
    )
    e.append_vertex(
        (16606.65151901649, 81.88147523282441, 2099.9999999999995),
        dxfattribs={"flags": 64},
    )
    e.append_vertex(
        (16606.65151901649, 81.88147523282441, 1199.9999999999998),
        dxfattribs={"flags": 64},
    )
    e.append_vertex(
        (16606.65151901649, 81.88147523282441, 1199.9999999999998),
        dxfattribs={"flags": 64},
    )
    e.append_vertex(
        (16606.65151901649, 1281.8814752328244, 2099.9999999999995),
        dxfattribs={"flags": 64},
    )
    e.append_vertex(
        (16606.65151901649, 1281.8814752328244, 1199.9999999999998),
        dxfattribs={"flags": 64},
    )
    e.append_vertex(
        (16606.65151901649, 1281.8814752328244, 1199.9999999999998),
        dxfattribs={"flags": 64},
    )
    e.append_vertex(
        (16626.65151901649, 1281.8814752328244, 2099.9999999999995),
        dxfattribs={"flags": 64},
    )
    e.append_vertex(
        (16626.65151901649, 1281.8814752328244, 1199.9999999999998),
        dxfattribs={"flags": 64},
    )
    e.append_vertex(
        (16606.65151901649, 1281.8814752328244, 1199.9999999999998),
        dxfattribs={"flags": 64},
    )
    e.append_vertex(
        (16626.65151901649, 81.88147523282441, 2099.9999999999995),
        dxfattribs={"flags": 64},
    )
    e.append_vertex(
        (16626.65151901649, 81.88147523282441, 1199.9999999999998),
        dxfattribs={"flags": 64},
    )
    e.append_vertex(
        (16606.65151901649, 81.88147523282441, 1199.9999999999998),
        dxfattribs={"flags": 64},
    )
    e.append_vertex(
        (16626.65151901649, 81.88147523282441, 2099.9999999999995),
        dxfattribs={"flags": 64},
    )
    e.append_vertex(
        (16626.65151901649, 81.88147523282441, 2099.9999999999995),
        dxfattribs={"flags": 64},
    )
    e.append_vertex(
        (16606.65151901649, 81.88147523282441, 2099.9999999999995),
        dxfattribs={"flags": 64},
    )
    e.append_vertex(
        (16626.65151901649, 1281.8814752328244, 2099.9999999999995),
        dxfattribs={"flags": 64},
    )
    e.append_vertex(
        (16626.65151901649, 1281.8814752328244, 2099.9999999999995),
        dxfattribs={"flags": 64},
    )
    e.append_vertex(
        (16606.65151901649, 1281.8814752328244, 2099.9999999999995),
        dxfattribs={"flags": 64},
    )
    return e


def test_from_polyface_182_2(polyface_181_2):
    mesh = MeshVertexMerger.from_polyface(polyface_181_2)
    assert len(mesh.vertices) == 8


def test_mesh_subdivide():
    c = forms.cube().scale_uniform(10).subdivide(2)
    assert len(c.vertices) == 2 * 25 + 3 * 16
    assert len(c.faces) == 16 * 6


def test_debug_coplanar_faces():
    source = MeshBuilder()
    source.vertices = [
        Vec3(-5.0, -5.0, -5.0),  # 0
        Vec3(-5.0, 0.0, -5.0),  # 1
        Vec3(0.0, 0.0, -5.0),  # 2
        Vec3(0.0, -5.0, -5.0),  # 3
        Vec3(-5.0, 5.0, -5.0),  # 4
        Vec3(0.0, 5.0, -5.0),  # 5
        Vec3(5.0, 5.0, -5.0),  # 6
        Vec3(5.0, 0.0, -5.0),  # 7
        Vec3(5.0, -5.0, -5.0),  # 8
    ]
    source.faces = [
        (0, 1, 2, 3),
        (4, 5, 2, 1),
        (6, 7, 2, 5),
        (8, 3, 2, 7),
    ]
    optimized_cube = source.merge_coplanar_faces()
    assert len(optimized_cube.faces) == 1
    assert len(optimized_cube.vertices) == 4


def test_merge_coplanar_faces():
    c = forms.cube().scale_uniform(10).subdivide(1)
    assert len(c.vertices) == 26
    assert len(c.faces) == 24
    optimized_cube = c.merge_coplanar_faces()
    assert len(optimized_cube.faces) == 6
    assert len(optimized_cube.vertices) == 8


def test_merge_disk():
    m = MeshVertexMerger()
    vertices = list(forms.circle(8, close=True))
    for v1, v2 in zip(vertices, vertices[1:]):
        m.add_face([Vec3(), v1, v2])
    assert len(m.vertices) == 9
    assert len(m.faces) == 8

    m2 = m.merge_coplanar_faces()
    assert len(m2.vertices) == 8
    assert len(m2.faces) == 1


def test_merge_coplanar_faces_in_two_passes():
    c = forms.cube().scale_uniform(10).subdivide(2)
    assert len(c.vertices) == 98
    assert len(c.faces) == 96
    optimized_cube = c.merge_coplanar_faces(passes=2)
    assert len(optimized_cube.faces) == 6
    assert len(optimized_cube.vertices) == 8


class TestMergeConnectedPaths:
    @pytest.mark.parametrize(
        "p",
        [
            [1, 2, 3],
            [3, 2, 1],
            [1, 2, 3, 4, 5],
        ],
    )
    def test_non_connected_paths(self, p):
        assert merge_connected_paths(p, [17, 18, 19]) == p

    def test_connected_squares_same_orientation(self):
        # fmt: off
        assert merge_connected_paths([1, 2, 3, 4], [4, 3, 5, 6]) == [
            1, 2, 3, 5, 6, 4,
        ]
        assert merge_connected_paths([1, 2, 3, 4], [4, 5, 6, 1]) == [
            1, 2, 3, 4, 5, 6
        ]
        # fmt: on

    def test_connected_squares_different_orientation(self):
        """The connected structure have to have the same orientation (clockwise
        or counter-clockwise to be merged.
        """
        # fmt: off
        assert merge_connected_paths([1, 2, 3, 4], [6, 5, 3, 4]) == [
            1, 2, 3, 4
        ]
        # fmt: on

    def test_connected_rect_same_orientation(self):
        # fmt: off
        assert merge_connected_paths([1, 2, 3, 4, 5, 6], [6, 5, 4, 7, 8, 9]) == [
            1, 2, 3, 4, 7, 8, 9, 6
        ]
        # fmt: on

    def test_complex_shape(self):
        # fmt: off
        assert merge_connected_paths(
            [1, 2, 3, 4, 5, 6, 7, 8],
            [6, 9, 10, 11, 8, 7],
        ) == [
            1, 2, 3, 4, 5, 6, 9, 10, 11, 8,
        ]
        # fmt: on

    def test_connected_by_one_vertex(self):
        with pytest.raises(NodeMergingError):
            merge_connected_paths([1, 2, 3, 4], [5, 6, 7, 4])

    @pytest.mark.parametrize(
        "p2",
        [
            [5, 6, 7, 4],
            [5, 6, 7, 4, 8],
            [5, 6, 7, 4, 8, 9],
        ],
    )
    def test_connection_error(self, p2):
        with pytest.raises(NodeMergingError):
            merge_connected_paths([1, 2, 3, 4], p2)

    def test_merge_multiple_paths(self):
        p1 = [1, 2, 9, 8]
        p2 = [3, 4, 9, 2]
        p3 = [5, 6, 9, 4]
        p4 = [7, 8, 9, 6]
        p = merge_connected_paths(p1, p2)
        assert p == [1, 2, 3, 4, 9, 8]
        p = merge_connected_paths(p, p3)
        assert p == [1, 2, 3, 4, 5, 6, 9, 8]
        p = merge_connected_paths(p, p4)
        assert p == [1, 2, 3, 4, 5, 6, 7, 8]

    def test_degenerated_path(self):
        """This creates a path [0, 1] which is invalid."""
        open_segments = [0, 1, 2, 3, 4, 5, 6, 7, 8]
        closing_segment = [0, 8, 1]
        with pytest.raises(DegeneratedPathError):
            merge_connected_paths(open_segments, closing_segment)


class TestRemoveColinearVertices:
    @pytest.mark.parametrize(
        "v",
        [
            [],
            [Vec3(0, 0)],
            [Vec3(0, 0), Vec3(1, 0)],
        ],
    )
    def test_simple_cases_without_action(self, v):
        assert list(remove_colinear_face_vertices(v)) == v

    @pytest.mark.parametrize(
        "v",
        [
            [Vec3(0, 0), Vec3(1, 0), Vec3(2, 0)],
            [Vec3(0, 0), Vec3(1, 0), Vec3(2, 0), Vec3(3, 0)],
        ],
    )
    def test_remove_in_between_vertices(self, v):
        assert list(remove_colinear_face_vertices(v)) == [v[0], v[-1]]

    @pytest.mark.parametrize(
        "v",
        [
            [Vec3(0, 0), Vec3(0, 0), Vec3(2, 0)],
            [Vec3(0, 0), Vec3(0, 0), Vec3(0, 0), Vec3(3, 0)],
            [Vec3(0, 0), Vec3(0, 0), Vec3(0, 0), Vec3(0, 0)],
            [Vec3(1, 0), Vec3(1, 0), Vec3(0, 2), Vec3(0, 2)],
        ],
    )
    def test_remove_duplicated_vertices(self, v):
        assert list(remove_colinear_face_vertices(v)) == [v[0], v[-1]]

    def test_remove_in_between_vertices_with_direction_change(self):
        v = [
            Vec3(0, 0),
            Vec3(1, 0),
            Vec3(2, 0),
            Vec3(2, 1),
            Vec3(2, 2),
            Vec3(2, 3),
        ]
        assert list(remove_colinear_face_vertices(v)) == [v[0], v[2], v[-1]]

    def test_subdivided_square(self):
        v = [
            Vec3(-5.0, -5.0, -5.0),
            Vec3(-5.0, 0.0, -5.0),
            Vec3(-5.0, 5.0, -5.0),
            Vec3(0.0, 5.0, -5.0),
            Vec3(5.0, 5.0, -5.0),
            Vec3(5.0, 0.0, -5.0),
            Vec3(5.0, -5.0, -5.0),
            Vec3(0.0, -5.0, -5.0),
        ]
        assert list(remove_colinear_face_vertices(v)) == [
            v[0],
            v[2],
            v[4],
            v[6],
        ]


class TestMergeFullPatch:
    @pytest.mark.parametrize(
        "seg",
        [
            [0, 8, 1],
            [1, 0, 8],
            [8, 1, 0],
        ],
    )
    def test_fill_pie(self, seg):
        open_pie = [0, 1, 2, 3, 4, 5, 6, 7, 8]
        res = merge_full_patch(open_pie, seg)
        assert res == [1, 2, 3, 4, 5, 6, 7, 8]


def test_all_edges_cube():
    mesh = forms.cube()
    edges = list(all_edges(mesh.faces))
    assert len(edges) == 6 * 4
    assert len(set(edges)) == 24


class TestGetEdgeStats:
    @pytest.fixture(scope="class")
    def edges(self):
        mesh = forms.cube()
        return get_edge_stats(mesh.faces)

    def test_unique_edge_count(self, edges):
        assert len(edges) == 12

    def test_sum_of_edge_count(self, edges):
        assert sum(e[0] for e in edges.values()) == 24

    def test_all_balances_are_0(self, edges):
        assert all(e[1] == 0 for e in edges.values()) is True

    def test_invalid_face_orientation_break_the_rules(self):
        faces = forms.cube().faces
        faces[-1] = list(reversed(faces[-1]))
        edges = get_edge_stats(faces)
        assert all(e[1] == 0 for e in edges.values()) is False

    def test_coincident_faces_break_the_rules(self):
        faces = forms.cube().faces
        faces.append(faces[-1])
        edges = get_edge_stats(faces)
        assert all(e[1] == 0 for e in edges.values()) is False

    def test_edge_balance_has_no_meaning_for_open_surfaces(self):
        faces = [(0, 1, 2)]
        edges = get_edge_stats(faces)
        assert all(e[1] != 0 for e in edges.values()) is True


class TestSeparateMeshes:
    def test_separate_a_single_cube_returns_a_single_cube(self):
        c1 = forms.cube()
        result = list(separate_meshes(c1))
        assert len(result) == 1
        c2 = result[0]
        # vertex and face structure keeps stable for single meshes:
        assert c1.vertices == c2.vertices
        assert c1.faces == c2.faces

    def test_separate_menger_sponge(self):
        m1 = MengerSponge().mesh()
        result = list(separate_meshes(m1))
        assert len(result) == 1
        m2 = result[0]
        # vertex and face structure keeps stable for single meshes:
        assert m1.vertices == m2.vertices
        assert m1.faces == m2.faces

    def test_separate_two_cubes(self):
        cubes = forms.cube()
        cubes.translate(10, 0, 0)
        cubes.add_mesh(mesh=forms.cube())
        # a non-broken edge balance is a requirement to work properly:
        assert cubes.diagnose().is_edge_balance_broken is False
        result = list(separate_meshes(cubes))
        assert len(result) == 2

    def test_separate_two_intersecting_cubes(self):
        cubes = forms.cube()
        cubes.translate(0.2, 0.2, 0.2)
        cubes.add_mesh(mesh=forms.cube())
        # a non-broken edge balance is a requirement to work properly:
        assert cubes.diagnose().is_edge_balance_broken is False
        assert len(cubes.separate_meshes()) == 2


class TestNormals:
    @pytest.fixture(scope="class")
    def normals(self):
        return list(forms.cube().face_normals())

    def test_cube_has_six_normals(self, normals):
        assert len(normals) == 6

    def test_all_normals_are_normalized(self, normals):
        assert all(math.isclose(n.magnitude, 1.0) for n in normals) is True

    def test_all_normals_are_different(self, normals):
        assert len(set(normals)) == 6


def test_concave_mesh_tessellation():
    profile = [(0, 0), (10, 0), (10, 10), (8, 10), (8, 2), (0, 2)]
    concave_prism = forms.extrude(
        profile, [(0, 0, 0), (0, 0, 10)], close=True, caps=True
    )
    diag = concave_prism.mesh_tessellation(3).diagnose()
    assert all(len(face) == 3 for face in diag.faces) is True
    assert diag.n_vertices == 12
    assert diag.n_faces == 20
    assert diag.n_edges == 30
    assert diag.is_manifold is True


@pytest.mark.parametrize(
    "sx,sy,sz,expected",
    [
        (1, 1, 1, True),
        (-1, 1, 1, False),
        (1, -1, 1, False),
        (1, 1, -1, False),
        (-1, -1, 1, True),
        (-1, 1, -1, True),
        (1, -1, -1, True),
        (-1, -1, -1, False),
    ],
    ids="none x y z xy xz yz xyz".split(),
)
def test_check_face_normals_after_transformation(sx, sy, sz, expected):
    assert face_normals_after_transformation(Matrix44.scale(sx, sy, sz)) is expected


class TestFaceOrientationDetector:
    def test_cube_has_uniform_face_normals(self):
        fod = FaceOrientationDetector(forms.cube())
        assert fod.has_uniform_face_normals is True
        assert fod.count == (6, 0)
        assert fod.is_manifold is True
        assert fod.all_reachable is True

    def test_multiple_disconnected_cubes(self):
        """A MeshBuilder can contain multiple disconnected meshes.
        Only the mesh which contains the reference face will be processed.

        The attribute all_reachable shows if all faces from the MeshBuilder are
        processed.
        """
        cubes = forms.cube()
        cube1 = forms.cube()
        cube1.translate(10, 0, 0)
        cubes.add_mesh(mesh=cube1)

        fod = FaceOrientationDetector(cubes)
        assert fod.has_uniform_face_normals is True
        assert fod.count == (6, 0)
        assert fod.is_manifold is True
        assert (
            fod.all_reachable is False
        ), "not all faces are reachable from the reference face"

    def test_flipped_cube_faces_are_also_uniform(self):
        """It's always the reference face which determines the forward
        orientation! In this case all face normals are pointing inwards, which
        is also a uniform orientation.
        """
        cube = forms.cube()
        cube.flip_normals()
        fod = FaceOrientationDetector(cube)
        assert fod.has_uniform_face_normals is True
        assert fod.count == (6, 0)
        assert fod.is_manifold is True
        assert fod.all_reachable is True

    def test_modified_cube_has_not_uniform_face_normals(self):
        cube = forms.cube()
        cube.faces[-1] = tuple(reversed(cube.faces[-1]))
        fod = FaceOrientationDetector(cube)
        assert fod.has_uniform_face_normals is False
        assert list(fod.backward_faces)[0] == cube.faces[-1]
        assert fod.count == (5, 1)
        assert fod.is_manifold is True
        assert fod.all_reachable is True

    def test_torus_with_uniform_face_normals(self):
        fod = FaceOrientationDetector(forms.torus())
        assert fod.has_uniform_face_normals is True
        assert fod.count == (128, 0)
        assert fod.is_manifold is True
        assert fod.all_reachable is True

    def test_find_all_backward_oriented_faces(self):
        torus = forms.torus()
        # reverse the first face (= reference face)
        torus.faces[0] = tuple(reversed(torus.faces[0]))
        fod = FaceOrientationDetector(torus)
        assert fod.has_uniform_face_normals is False
        assert fod.count == (1, 127)
        assert fod.is_manifold is True
        assert fod.all_reachable is True

    def test_detect_reference_face_orientation(self):
        # Calculation is tested for is_face_normal_pointing_outwards() in 661!
        fod = FaceOrientationDetector(forms.cube())
        assert fod.is_reference_face_pointing_outwards() is True

    def test_regular_open_surface(self):
        cube = forms.cube()
        cube.faces.pop()
        fod = FaceOrientationDetector(cube)
        assert fod.is_closed_surface is False

    def test_regular_cube_has_closed_surface(self):
        fod = FaceOrientationDetector(forms.cube())
        assert fod.is_closed_surface is True

    def test_multiple_cubes_have_closed_surface(self):
        cube = forms.cube()
        cube2 = cube.copy()
        cube2.translate(10)
        cube.add_mesh(mesh=cube2)
        fod = FaceOrientationDetector(cube)
        assert fod.is_closed_surface is True

    def test_cube_with_reversed_face_has_closed_surface(self):
        cube = forms.cube()
        cube.faces[0] = tuple(reversed(cube.faces[0]))
        fod = FaceOrientationDetector(cube)
        assert fod.is_closed_surface is True

    def test_non_manifold_cube_has_not_a_closed_surface(self):
        cube = forms.cube()
        cube.faces.append(cube.faces[0])
        fod = FaceOrientationDetector(cube)
        assert fod.is_closed_surface is False


def test_unify_cube_normals_by_reference_face():
    cube = forms.cube()
    cube.faces[-1] = tuple(reversed(cube.faces[-1]))
    cube2 = cube.unify_face_normals_by_reference()
    fod = FaceOrientationDetector(cube2)
    assert fod.has_uniform_face_normals is True
    assert fod.count == (6, 0)
    assert fod.is_manifold is True
    assert fod.all_reachable is True
    assert fod.is_reference_face_pointing_outwards() is True


@pytest.mark.parametrize("force", [False, True])
def test_force_unified_cube_normals_pointing_outwards(force):
    cube = forms.cube()
    cube.faces[-1] = tuple(reversed(cube.faces[-1]))
    cube2 = cube.unify_face_normals_by_reference(-1, force_outwards=force)
    fod = FaceOrientationDetector(cube2)
    assert fod.is_reference_face_pointing_outwards() is force


def test_unify_cube_normals_by_majority():
    cube = forms.cube()
    # reverse the first face (= reference face)
    cube.faces[0] = tuple(reversed(cube.faces[0]))
    cube2 = cube.unify_face_normals()
    fod = FaceOrientationDetector(cube2)
    assert fod.has_uniform_face_normals is True
    assert fod.count == (6, 0)
    assert fod.is_manifold is True
    assert fod.all_reachable is True


def test_unify_torus_normals_by_majority():
    torus = forms.torus()
    faces = torus.faces
    # remove some faces
    faces = [
        f for i, f in enumerate(faces) if i not in {2, 3, 17, 34, 99, 100, 101, 120}
    ]
    # reverse some face normals
    for i in [0, 3, 17, 34, 99, 100, 101, 119]:
        faces[i] = tuple(reversed(faces[i]))
    torus.faces = faces

    # unify face normals
    unified_torus = torus.unify_face_normals()
    fod = FaceOrientationDetector(unified_torus)
    assert fod.has_uniform_face_normals is True
    assert fod.count == (120, 0)
    assert fod.is_manifold is True
    assert fod.all_reachable is True


def test_volume6():
    assert volume6(Vec3(0, 0, 1), Vec3(2, 0, 1), Vec3(1, 1, 1)) == pytest.approx(2.0)


def test_volume_of_closed_surface():
    diag = forms.cube().diagnose()
    assert diag.volume() == pytest.approx(1.0)


def test_volume_of_surface_with_clockwise_oriented_vertices():
    cube = forms.cube()
    cube.flip_normals()
    assert cube.diagnose().volume() == pytest.approx(-1.0)


@pytest.mark.parametrize(
    "loc",
    [
        (100, 100, 100),
        (-100, -100, -100),
    ],
)
def test_location_of_mesh_is_not_relevant_for_volume_calculation(loc):
    cube = forms.cube()
    cube.translate(loc)
    assert cube.diagnose().volume() == pytest.approx(1.0)


def test_volume_of_an_open_surface_is_null():
    cube = forms.cube()
    del cube.faces[-1]
    assert cube.diagnose().volume() == 0.0


def test_volume_of_two_cubes_in_a_single_mesh_is_invalid():
    cube = forms.cube()
    c1 = forms.cube()
    c1.translate(2, 0, 0)
    cube.add_mesh(mesh=c1)
    assert len(cube.vertices) == 16
    assert len(cube.faces) == 12

    diag = forms.cube().diagnose()
    assert diag.volume() == pytest.approx(1.0)  # invalid result!!!


def test_torus_volume():
    r0 = 1.0
    r1 = 10.0
    real_volume = 2.0 * math.pi**2 * r0**2 * r1
    diag = forms.torus(
        major_count=64, minor_count=32, minor_radius=r0, major_radius=r1
    ).diagnose()
    # less than 2% difference expected:
    assert real_volume * 0.98 < diag.volume() < real_volume


def test_surface_area_of_a_cube():
    diag = forms.cube().diagnose()
    assert diag.surface_area() == pytest.approx(6.0)


def test_surface_area_of_a_torus():
    r0 = 1.0
    r1 = 10.0
    real_area = 4.0 * math.pi**2 * r0 * r1
    diag = forms.torus(
        major_count=64, minor_count=32, minor_radius=r0, major_radius=r1
    ).diagnose()
    # less than 2% difference expected:
    assert real_area * 0.98 < diag.surface_area() < real_area


def test_centroid_of_a_cube():
    diag = forms.cube(center=False).diagnose()
    assert diag.centroid().isclose((0.5, 0.5, 0.5))
